#!/usr/bin/env python
# -*- coding: utf-8 -*-
#*****************************************************************************
# $Id$
#
# Project:  OpenGIS Simple Features Reference Implementation
# Purpose:  Python port of a simple client for translating between formats.
# Author:   Even Rouault, <even dot rouault at mines dash paris dot org>
#
# Port from ogr2ogr.cpp whose author is Frank Warmerdam
#
#*****************************************************************************
# Copyright (c) 2010-2013, Even Rouault <even dot rouault at mines-paris dot org>
# Copyright (c) 1999, Frank Warmerdam
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#**************************************************************************

# Note : this is the most direct port of ogr2ogr.cpp possible
# It could be made much more Python'ish !

import sys
import os
import stat
import helpfunctions as hf

from osgeo import gdal
from osgeo import ogr
from osgeo import osr

###############################################################################

class ScaledProgressObject:
    def __init__(self, min, max, cbk, cbk_data = None):
        self.min = min
        self.max = max
        self.cbk = cbk
        self.cbk_data = cbk_data

###############################################################################

def ScaledProgressFunc(pct, msg, data):
    if data.cbk is None:
        return True
    return data.cbk(data.min + pct * (data.max - data.min), msg, data.cbk_data)

###############################################################################

def EQUAL(a, b):
    return a.lower() == b.lower()

###############################################################################
# Redefinition of GDALTermProgress, so that autotest/pyscripts/test_ogr2ogr_py.py
# can check that the progress bar is displayed

nLastTick = -1

def TermProgress( dfComplete, pszMessage, pProgressArg ):

    global nLastTick
    nThisTick = (int) (dfComplete * 40.0)

    if nThisTick < 0:
        nThisTick = 0
    if nThisTick > 40:
        nThisTick = 40

    # Have we started a new progress run?
    if nThisTick < nLastTick and nLastTick >= 39:
        nLastTick = -1

    if nThisTick <= nLastTick:
        return True

    while nThisTick > nLastTick:
        nLastTick = nLastTick + 1
        if (nLastTick % 4) == 0:
            sys.stdout.write('%d' % ((nLastTick / 4) * 10))
        else:
            sys.stdout.write('.')

    if nThisTick == 40:
        print(" - done." )
    else:
        sys.stdout.flush()

    return True

class TargetLayerInfo:
    def __init__(self):
        self.poDstLayer = None
        self.poCT = None
        #self.papszTransformOptions = None
        self.panMap = None
        self.iSrcZField = None

class AssociatedLayers:
    def __init__(self):
        self.poSrcLayer = None
        self.psInfo = None

#**********************************************************************
#                                main()
#**********************************************************************

bSkipFailures = False
nGroupTransactions = 200
bPreserveFID = False
nFIDToFetch = ogr.NullFID

class Enum(set):
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

GeomOperation = Enum(["NONE", "SEGMENTIZE", "SIMPLIFY_PRESERVE_TOPOLOGY"])

def main(args = None, progress_func = TermProgress, progress_data = None):

    global bSkipFailures
    global nGroupTransactions
    global bPreserveFID
    global nFIDToFetch

    pszFormat = "ESRI Shapefile"
    pszDataSource = None
    pszDestDataSource = None
    papszLayers = []
    papszDSCO = []
    papszLCO = []
    bTransform = False
    bAppend = False
    bUpdate = False
    bOverwrite = False
    pszOutputSRSDef = None
    pszSourceSRSDef = None
    poOutputSRS = None
    bNullifyOutputSRS = False
    poSourceSRS = None
    pszNewLayerName = None
    pszWHERE = None
    poSpatialFilter = None
    pszSelect = None
    papszSelFields = None
    pszSQLStatement = None
    eGType = -2
    bPromoteToMulti = False
    eGeomOp = GeomOperation.NONE
    dfGeomOpParam = 0
    papszFieldTypesToString = []
    bDisplayProgress = False
    pfnProgress = None
    pProgressArg = None
    bClipSrc = False
    bWrapDateline = False
    poClipSrc = None
    pszClipSrcDS = None
    pszClipSrcSQL = None
    pszClipSrcLayer = None
    pszClipSrcWhere = None
    poClipDst = None
    pszClipDstDS = None
    pszClipDstSQL = None
    pszClipDstLayer = None
    pszClipDstWhere = None
    #pszSrcEncoding = None
    #pszDstEncoding = None
    bWrapDateline = False
    bExplodeCollections = False
    pszZField = None
    nCoordDim = -1

    if args is None:
        args = sys.argv

    args = ogr.GeneralCmdLineProcessor( args )

# --------------------------------------------------------------------
#      Processing command line arguments.
# --------------------------------------------------------------------
    if args is None:
        return False

    nArgc = len(args)

    iArg = 1
    while iArg < nArgc:
        if EQUAL(args[iArg],"-f") and iArg < nArgc-1:
            iArg = iArg + 1
            pszFormat = args[iArg]

        elif EQUAL(args[iArg],"-dsco") and iArg < nArgc-1:
            iArg = iArg + 1
            papszDSCO.append(args[iArg] )

        elif EQUAL(args[iArg],"-lco") and iArg < nArgc-1:
            iArg = iArg + 1
            papszLCO.append(args[iArg] )

        elif EQUAL(args[iArg],"-preserve_fid"):
            bPreserveFID = True

        elif len(args[iArg]) >= 5 and EQUAL(args[iArg][0:5], "-skip"):
            bSkipFailures = True
            nGroupTransactions = 1 # #2409

        elif EQUAL(args[iArg],"-append"):
            bAppend = True
            bUpdate = True

        elif EQUAL(args[iArg],"-overwrite"):
            bOverwrite = True
            bUpdate = True

        elif EQUAL(args[iArg],"-update"):
            bUpdate = True

        elif EQUAL(args[iArg],"-fid") and iArg < nArgc-1:
            iArg = iArg + 1
            nFIDToFetch = int(args[iArg])

        elif EQUAL(args[iArg],"-sql") and iArg < nArgc-1:
            iArg = iArg + 1
            pszSQLStatement = args[iArg]

        elif EQUAL(args[iArg],"-nln") and iArg < nArgc-1:
            iArg = iArg + 1
            pszNewLayerName = args[iArg]

        elif EQUAL(args[iArg],"-nlt") and iArg < nArgc-1:

            if EQUAL(args[iArg+1],"NONE"):
                eGType = ogr.wkbNone
            elif EQUAL(args[iArg+1],"GEOMETRY"):
                eGType = ogr.wkbUnknown
            elif EQUAL(args[iArg+1],"PROMOTE_TO_MULTI"):
                bPromoteToMulti = True
            elif EQUAL(args[iArg+1],"POINT"):
                eGType = ogr.wkbPoint
            elif EQUAL(args[iArg+1],"LINESTRING"):
                eGType = ogr.wkbLineString
            elif EQUAL(args[iArg+1],"POLYGON"):
                eGType = ogr.wkbPolygon
            elif EQUAL(args[iArg+1],"GEOMETRYCOLLECTION"):
                eGType = ogr.wkbGeometryCollection
            elif EQUAL(args[iArg+1],"MULTIPOINT"):
                eGType = ogr.wkbMultiPoint
            elif EQUAL(args[iArg+1],"MULTILINESTRING"):
                eGType = ogr.wkbMultiLineString
            elif EQUAL(args[iArg+1],"MULTIPOLYGON"):
                eGType = ogr.wkbMultiPolygon
            elif EQUAL(args[iArg+1],"GEOMETRY25D"):
                eGType = ogr.wkbUnknown | ogr.wkb25DBit
            elif EQUAL(args[iArg+1],"POINT25D"):
                eGType = ogr.wkbPoint25D
            elif EQUAL(args[iArg+1],"LINESTRING25D"):
                eGType = ogr.wkbLineString25D
            elif EQUAL(args[iArg+1],"POLYGON25D"):
                eGType = ogr.wkbPolygon25D
            elif EQUAL(args[iArg+1],"GEOMETRYCOLLECTION25D"):
                eGType = ogr.wkbGeometryCollection25D
            elif EQUAL(args[iArg+1],"MULTIPOINT25D"):
                eGType = ogr.wkbMultiPoint25D
            elif EQUAL(args[iArg+1],"MULTILINESTRING25D"):
                eGType = ogr.wkbMultiLineString25D
            elif EQUAL(args[iArg+1],"MULTIPOLYGON25D"):
                eGType = ogr.wkbMultiPolygon25D
            else:
                print("-nlt %s: type not recognised." % args[iArg+1])
                return False

            iArg = iArg + 1

        elif EQUAL(args[iArg],"-dim") and iArg < nArgc-1:

            nCoordDim = int(args[iArg+1])
            if nCoordDim != 2 and nCoordDim != 3:
                print("-dim %s: value not handled." % args[iArg+1])
                return False
            iArg = iArg + 1

        elif (EQUAL(args[iArg],"-tg") or \
                EQUAL(args[iArg],"-gt")) and iArg < nArgc-1:
            iArg = iArg + 1
            nGroupTransactions = int(args[iArg])

        elif EQUAL(args[iArg],"-s_srs") and iArg < nArgc-1:
            iArg = iArg + 1
            pszSourceSRSDef = args[iArg]

        elif EQUAL(args[iArg],"-a_srs") and iArg < nArgc-1:
            iArg = iArg + 1
            pszOutputSRSDef = args[iArg]
            if EQUAL(pszOutputSRSDef, "NULL") or \
               EQUAL(pszOutputSRSDef, "NONE"):
                pszOutputSRSDef = None
                bNullifyOutputSRS = True

        elif EQUAL(args[iArg],"-t_srs") and iArg < nArgc-1:
            iArg = iArg + 1
            pszOutputSRSDef = args[iArg]
            bTransform = True

        elif EQUAL(args[iArg],"-spat") and iArg + 4 < nArgc:
            oRing = ogr.Geometry(ogr.wkbLinearRing)

            oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+2]) )
            oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+4]) )
            oRing.AddPoint_2D( float(args[iArg+3]), float(args[iArg+4]) )
            oRing.AddPoint_2D( float(args[iArg+3]), float(args[iArg+2]) )
            oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+2]) )

            poSpatialFilter = ogr.Geometry(ogr.wkbPolygon)
            poSpatialFilter.AddGeometry(oRing)
            iArg = iArg + 4

        elif EQUAL(args[iArg],"-where") and iArg < nArgc-1:
            iArg = iArg + 1
            pszWHERE = args[iArg]

        elif EQUAL(args[iArg],"-select") and iArg < nArgc-1:
            iArg = iArg + 1
            pszSelect = args[iArg]
            if pszSelect.find(',') != -1:
                papszSelFields = pszSelect.split(',')
            else:
                papszSelFields = pszSelect.split(' ')
            if papszSelFields[0] == '':
                papszSelFields = []

        elif EQUAL(args[iArg],"-simplify") and iArg < nArgc-1:
            iArg = iArg + 1
            eGeomOp = GeomOperation.SIMPLIFY_PRESERVE_TOPOLOGY
            dfGeomOpParam = float(args[iArg])

        elif EQUAL(args[iArg],"-segmentize") and iArg < nArgc-1:
            iArg = iArg + 1
            eGeomOp = GeomOperation.SEGMENTIZE
            dfGeomOpParam = float(args[iArg])

        elif EQUAL(args[iArg],"-fieldTypeToString") and iArg < nArgc-1:
            iArg = iArg + 1
            pszFieldTypeToString = args[iArg]
            if pszFieldTypeToString.find(',') != -1:
                tokens = pszFieldTypeToString.split(',')
            else:
                tokens = pszFieldTypeToString.split(' ')

            for token in tokens:
                if EQUAL(token,"Integer") or \
                    EQUAL(token,"Real") or \
                    EQUAL(token,"String") or \
                    EQUAL(token,"Date") or \
                    EQUAL(token,"Time") or \
                    EQUAL(token,"DateTime") or \
                    EQUAL(token,"Binary") or \
                    EQUAL(token,"IntegerList") or \
                    EQUAL(token,"RealList") or \
                    EQUAL(token,"StringList"):

                    papszFieldTypesToString.append(token)

                elif EQUAL(token,"All"):
                    papszFieldTypesToString = [ 'All' ]
                    break

                else:
                    print("Unhandled type for fieldtypeasstring option : %s " % token)
                    return Usage()

        elif EQUAL(args[iArg],"-progress"):
            bDisplayProgress = True

        #elif EQUAL(args[iArg],"-wrapdateline") )
        #{
        #    bWrapDateline = True;
        #}
        #
        elif EQUAL(args[iArg],"-clipsrc") and iArg < nArgc-1:

            bClipSrc = True
            if IsNumber(args[iArg+1]) and iArg < nArgc - 4:
                oRing = ogr.Geometry(ogr.wkbLinearRing)

                oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+2]) )
                oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+4]) )
                oRing.AddPoint_2D( float(args[iArg+3]), float(args[iArg+4]) )
                oRing.AddPoint_2D( float(args[iArg+3]), float(args[iArg+2]) )
                oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+2]) )

                poClipSrc = ogr.Geometry(ogr.wkbPolygon)
                poClipSrc.AddGeometry(oRing)
                iArg = iArg + 4

            elif (len(args[iArg+1]) >= 7 and EQUAL(args[iArg+1][0:7],"POLYGON") ) or \
                  (len(args[iArg+1]) >= 12 and EQUAL(args[iArg+1][0:12],"MULTIPOLYGON") ) :
                poClipSrc = ogr.CreateGeometryFromWkt(args[iArg+1])
                if poClipSrc is None:
                    print("FAILURE: Invalid geometry. Must be a valid POLYGON or MULTIPOLYGON WKT\n")
                    return Usage()

                iArg = iArg + 1

            elif EQUAL(args[iArg+1],"spat_extent"):
                iArg = iArg + 1

            else:
                pszClipSrcDS = args[iArg+1]
                iArg = iArg + 1

        elif EQUAL(args[iArg],"-clipsrcsql") and iArg < nArgc-1:
            pszClipSrcSQL = args[iArg+1]
            iArg = iArg + 1

        elif EQUAL(args[iArg],"-clipsrclayer") and iArg < nArgc-1:
            pszClipSrcLayer = args[iArg+1]
            iArg = iArg + 1

        elif EQUAL(args[iArg],"-clipsrcwhere") and iArg < nArgc-1:
            pszClipSrcWhere = args[iArg+1]
            iArg = iArg + 1

        elif EQUAL(args[iArg],"-clipdst") and iArg < nArgc-1:

            if IsNumber(args[iArg+1]) and iArg < nArgc - 4:
                oRing = ogr.Geometry(ogr.wkbLinearRing)

                oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+2]) )
                oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+4]) )
                oRing.AddPoint_2D( float(args[iArg+3]), float(args[iArg+4]) )
                oRing.AddPoint_2D( float(args[iArg+3]), float(args[iArg+2]) )
                oRing.AddPoint_2D( float(args[iArg+1]), float(args[iArg+2]) )

                poClipDst = ogr.Geometry(ogr.wkbPolygon)
                poClipDst.AddGeometry(oRing)
                iArg = iArg + 4

            elif (len(args[iArg+1]) >= 7 and EQUAL(args[iArg+1][0:7],"POLYGON") ) or \
                  (len(args[iArg+1]) >= 12 and EQUAL(args[iArg+1][0:12],"MULTIPOLYGON") ) :
                poClipDst = ogr.CreateGeometryFromWkt(args[iArg+1])
                if poClipDst is None:
                    print("FAILURE: Invalid geometry. Must be a valid POLYGON or MULTIPOLYGON WKT\n")
                    return Usage()

                iArg = iArg + 1

            elif EQUAL(args[iArg+1],"spat_extent"):
                iArg = iArg + 1

            else:
                pszClipDstDS = args[iArg+1]
                iArg = iArg + 1

        elif EQUAL(args[iArg],"-clipdstsql") and iArg < nArgc-1:
            pszClipDstSQL = args[iArg+1]
            iArg = iArg + 1

        elif EQUAL(args[iArg],"-clipdstlayer") and iArg < nArgc-1:
            pszClipDstLayer = args[iArg+1]
            iArg = iArg + 1

        elif EQUAL(args[iArg],"-clipdstwhere") and iArg < nArgc-1:
            pszClipDstWhere = args[iArg+1]
            iArg = iArg + 1

        elif EQUAL(args[iArg],"-explodecollections"):
            bExplodeCollections = True

        elif EQUAL(args[iArg],"-zfield") and iArg < nArgc-1:
            pszZField = args[iArg+1]
            iArg = iArg + 1

        elif args[iArg][0] == '-':
            return Usage()

        elif pszDestDataSource is None:
            pszDestDataSource = args[iArg]
        elif pszDataSource is None:
            pszDataSource = args[iArg]
        else:
            papszLayers.append (args[iArg] )

        iArg = iArg + 1

    if pszDataSource is None:
        return Usage()

    if bPreserveFID and bExplodeCollections:
        print("FAILURE: cannot use -preserve_fid and -explodecollections at the same time\n\n")
        return Usage()

    if bClipSrc and pszClipSrcDS is not None:
        poClipSrc = LoadGeometry(pszClipSrcDS, pszClipSrcSQL, pszClipSrcLayer, pszClipSrcWhere)
        if poClipSrc is None:
            print("FAILURE: cannot load source clip geometry\n" )
            return Usage()

    elif bClipSrc and poClipSrc is None:
        if poSpatialFilter is not None:
            poClipSrc = poSpatialFilter.Clone()
        if poClipSrc is None:
            print("FAILURE: -clipsrc must be used with -spat option or a\n" + \
                  "bounding box, WKT string or datasource must be specified\n")
            return Usage()

    if pszClipDstDS is not None:
        poClipDst = LoadGeometry(pszClipDstDS, pszClipDstSQL, pszClipDstLayer, pszClipDstWhere)
        if poClipDst is None:
            print("FAILURE: cannot load dest clip geometry\n" )
            return Usage()

# --------------------------------------------------------------------
#      Open data source.
# --------------------------------------------------------------------
    poDS = ogr.Open( pszDataSource, False )

# --------------------------------------------------------------------
#      Report failure
# --------------------------------------------------------------------
    if poDS is None:
        #print("FAILURE:\n" + \
                #"Unable to open datasource `%s' with the following drivers." % pszDataSource)
        pass
        for iDriver in range(ogr.GetDriverCount()):
            #print("  ->  " + ogr.GetDriver(iDriver).GetName() )
            pass
        return False

# --------------------------------------------------------------------
#      Try opening the output datasource as an existing, writable
# --------------------------------------------------------------------
    poODS = None
    poDriver = None

    if bUpdate:
        poODS = ogr.Open( pszDestDataSource, True )
        if poODS is None:

            if bOverwrite or bAppend:
                poODS = ogr.Open( pszDestDataSource, False )
                if poODS is None:
                    # the datasource doesn't exist at all
                    bUpdate = False
                else:
                    poODS.delete()
                    poODS = None

            if bUpdate:
                print("FAILURE:\n" +
                        "Unable to open existing output datasource `%s'." % pszDestDataSource)
                return False

        elif len(papszDSCO) > 0:
            print("WARNING: Datasource creation options ignored since an existing datasource\n" + \
                    "         being updated." )

        if poODS is not None:
            poDriver = poODS.GetDriver()

# --------------------------------------------------------------------
#      Find the output driver.
# --------------------------------------------------------------------
    if not bUpdate:
        poDriver = ogr.GetDriverByName(pszFormat)
        if poDriver is None:
            print("Unable to find driver `%s'." % pszFormat)
            print( "The following drivers are available:" )

            for iDriver in range(ogr.GetDriverCount()):
                print("  ->  %s" % ogr.GetDriver(iDriver).GetName() )

            return False

        if poDriver.TestCapability( ogr.ODrCCreateDataSource ) == False:
            print( "%s driver does not support data source creation." % pszFormat)
            return False

# --------------------------------------------------------------------
#      Special case to improve user experience when translating
#      a datasource with multiple layers into a shapefile. If the
#      user gives a target datasource with .shp and it does not exist,
#      the shapefile driver will try to create a file, but this is not
#      appropriate because here we have several layers, so create
#      a directory instead.
# --------------------------------------------------------------------
        if EQUAL(poDriver.GetName(), "ESRI Shapefile") and \
           pszSQLStatement is None and \
           (len(papszLayers) > 1 or \
            (len(papszLayers) == 0 and poDS.GetLayerCount() > 1)) and \
            pszNewLayerName is None and \
            EQUAL(os.path.splitext(pszDestDataSource)[1], ".SHP") :

            try:
                os.stat(pszDestDataSource)
            except:
                try:
                    # decimal 493 = octal 0755. Python 3 needs 0o755, but
                    # this syntax is only supported by Python >= 2.6
                    os.mkdir(pszDestDataSource, 493)
                except:
                    print("Failed to create directory %s\n"
                          "for shapefile datastore.\n" % pszDestDataSource )
                    return False

# --------------------------------------------------------------------
#      Create the output data source.
# --------------------------------------------------------------------
        poODS = poDriver.CreateDataSource( pszDestDataSource, options = papszDSCO )
        if poODS is None:
            print( "%s driver failed to create %s" % (pszFormat, pszDestDataSource ))
            return False

# --------------------------------------------------------------------
#      Parse the output SRS definition if possible.
# --------------------------------------------------------------------
    if pszOutputSRSDef is not None:
        poOutputSRS = osr.SpatialReference()
        if poOutputSRS.SetFromUserInput( pszOutputSRSDef ) != 0:
            print( "Failed to process SRS definition: %s" % pszOutputSRSDef )
            return False

# --------------------------------------------------------------------
#      Parse the source SRS definition if possible.
# --------------------------------------------------------------------
    if pszSourceSRSDef is not None:
        poSourceSRS = osr.SpatialReference()
        if poSourceSRS.SetFromUserInput( pszSourceSRSDef ) != 0:
            print( "Failed to process SRS definition: %s" % pszSourceSRSDef )
            return False

# --------------------------------------------------------------------
#      For OSM file.
# --------------------------------------------------------------------
    bSrcIsOSM = poDS.GetDriver() is not None and \
                             poDS.GetDriver().GetName() == "OSM"
    nSrcFileSize = 0
    if bSrcIsOSM and poDS.GetName() != "/vsistdin/":
        sStat = gdal.VSIStatL(poDS.GetName())
        if sStat is not None:
            nSrcFileSize = sStat.size

# --------------------------------------------------------------------
#      Special case for -sql clause.  No source layers required.
# --------------------------------------------------------------------
    if pszSQLStatement is not None:
        if pszWHERE is not None:
            print( "-where clause ignored in combination with -sql." )
        if len(papszLayers) > 0:
            print( "layer names ignored in combination with -sql." )

        poResultSet = poDS.ExecuteSQL( pszSQLStatement, poSpatialFilter, \
                                        None )

        if poResultSet is not None:
            nCountLayerFeatures = 0
            if bDisplayProgress:
                if bSrcIsOSM:
                    pfnProgress = progress_func
                    pProgressArg = progress_data

                elif not poResultSet.TestCapability(ogr.OLCFastFeatureCount):
                    print( "Progress turned off as fast feature count is not available.")
                    bDisplayProgress = False

                else:
                    nCountLayerFeatures = poResultSet.GetFeatureCount()
                    pfnProgress = progress_func
                    pProgressArg = progress_data

# --------------------------------------------------------------------
#      Special case to improve user experience when translating into
#      single file shapefile and source has only one layer, and that
#      the layer name isn't specified
# --------------------------------------------------------------------
            if EQUAL(poDriver.GetName(), "ESRI Shapefile") and \
                pszNewLayerName is None:
                try:
                    mode = os.stat(pszDestDataSource).st_mode
                    if (mode & stat.S_IFDIR) == 0:
                        pszNewLayerName = os.path.splitext(os.path.basename(pszDestDataSource))[0]
                except:
                    pass


            psInfo = SetupTargetLayer(  poDS, \
                                        poResultSet,
                                        poODS, \
                                        papszLCO, \
                                        pszNewLayerName, \
                                        bTransform, \
                                        poOutputSRS, \
                                        bNullifyOutputSRS, \
                                        poSourceSRS, \
                                        papszSelFields, \
                                        bAppend, eGType, bPromoteToMulti, nCoordDim, bOverwrite, \
                                        papszFieldTypesToString, \
                                        bWrapDateline, \
                                        bExplodeCollections, \
                                        pszZField, \
                                        pszWHERE )

            poResultSet.ResetReading()

            if psInfo is None or not TranslateLayer( psInfo, poDS, poResultSet, poODS, \
                                poOutputSRS, bNullifyOutputSRS, \
                                eGType, bPromoteToMulti, nCoordDim, \
                                eGeomOp, dfGeomOpParam, \
                                nCountLayerFeatures, \
                                poClipSrc, poClipDst, \
                                bExplodeCollections, \
                                nSrcFileSize, None, \
                                pfnProgress, pProgressArg ):
                print(
                        "Terminating translation prematurely after failed\n" + \
                        "translation from sql statement." )

                return False

            poDS.ReleaseResultSet( poResultSet )


# --------------------------------------------------------------------
#      Special case for layer interleaving mode.
# --------------------------------------------------------------------
    elif bSrcIsOSM and gdal.GetConfigOption("OGR_INTERLEAVED_READING", None) is None:

        gdal.SetConfigOption("OGR_INTERLEAVED_READING", "YES")

        #if (bSplitListFields)
        #{
        #    fprintf( stderr, "FAILURE: -splitlistfields not supported in this mode\n" );
        #    exit( 1 );
        #}

        nSrcLayerCount = poDS.GetLayerCount()
        pasAssocLayers = [ AssociatedLayers() for i in range(nSrcLayerCount) ]

# --------------------------------------------------------------------
#      Special case to improve user experience when translating into
#      single file shapefile and source has only one layer, and that
#      the layer name isn't specified
# --------------------------------------------------------------------

        if EQUAL(poDriver.GetName(), "ESRI Shapefile") and \
           (len(papszLayers) == 1 or nSrcLayerCount == 1) and pszNewLayerName is None:
            try:
                mode = os.stat(pszDestDataSource).st_mode
                if (mode & stat.S_IFDIR) == 0:
                    pszNewLayerName = os.path.splitext(os.path.basename(pszDestDataSource))[0]
            except:
                pass

        if bDisplayProgress and bSrcIsOSM:
            pfnProgress = progress_func
            pProgressArg = progress_data

# --------------------------------------------------------------------
#      If no target layer specified, use all source layers.
# --------------------------------------------------------------------
        if len(papszLayers) == 0:
            papszLayers = [ None for i in range(nSrcLayerCount) ]
            for iLayer in range(nSrcLayerCount):
                poLayer = poDS.GetLayer(iLayer)
                if poLayer is None:
                    print("FAILURE: Couldn't fetch advertised layer %d!" % iLayer)
                    return False

                papszLayers[iLayer] = poLayer.GetName()
        else:
            if bSrcIsOSM:
                osInterestLayers = "SET interest_layers ="
                for iLayer in range(len(papszLayers)):
                    if iLayer != 0:
                        osInterestLayers = osInterestLayers + ","
                    osInterestLayers = osInterestLayers + papszLayers[iLayer]

                poDS.ExecuteSQL(osInterestLayers, None, None)

# --------------------------------------------------------------------
#      First pass to set filters and create target layers.
# --------------------------------------------------------------------
        for iLayer in range(nSrcLayerCount):
            poLayer = poDS.GetLayer(iLayer)
            if poLayer is None:
                print("FAILURE: Couldn't fetch advertised layer %d!" % iLayer)
                return False

            pasAssocLayers[iLayer].poSrcLayer = poLayer

            if CSLFindString(papszLayers, poLayer.GetName()) >= 0:
                if pszWHERE is not None:
                    if poLayer.SetAttributeFilter( pszWHERE ) != 0:
                        print("FAILURE: SetAttributeFilter(%s) on layer '%s' failed.\n" % (pszWHERE, poLayer.GetName()) )
                        if not bSkipFailures:
                            return False

                if poSpatialFilter is not None:
                    poLayer.SetSpatialFilter( poSpatialFilter )

                psInfo = SetupTargetLayer( poDS, \
                                           poLayer, \
                                           poODS, \
                                           papszLCO, \
                                           pszNewLayerName, \
                                           bTransform, \
                                           poOutputSRS, \
                                           bNullifyOutputSRS, \
                                           poSourceSRS, \
                                           papszSelFields, \
                                           bAppend, eGType, bPromoteToMulti, nCoordDim, bOverwrite, \
                                           papszFieldTypesToString, \
                                           bWrapDateline, \
                                           bExplodeCollections, \
                                           pszZField, \
                                           pszWHERE )

                if psInfo is None and not bSkipFailures:
                    return False

                pasAssocLayers[iLayer].psInfo = psInfo
            else:
                pasAssocLayers[iLayer].psInfo = None

# --------------------------------------------------------------------
#      Second pass to process features in a interleaved layer mode.
# --------------------------------------------------------------------
        bHasLayersNonEmpty = True
        while bHasLayersNonEmpty:
            bHasLayersNonEmpty = False

            for iLayer in range(nSrcLayerCount):
                poLayer = pasAssocLayers[iLayer].poSrcLayer
                psInfo = pasAssocLayers[iLayer].psInfo
                anReadFeatureCount = [0]

                if psInfo is not None:
                    if not TranslateLayer(psInfo, poDS, poLayer, poODS, \
                                        poOutputSRS, bNullifyOutputSRS,  \
                                        eGType, bPromoteToMulti, nCoordDim, \
                                        eGeomOp, dfGeomOpParam,  \
                                        0,  \
                                        poClipSrc, poClipDst,  \
                                        bExplodeCollections,  \
                                        nSrcFileSize,  \
                                        anReadFeatureCount, \
                                        pfnProgress, pProgressArg ) \
                        and not bSkipFailures:
                        print(
                                "Terminating translation prematurely after failed\n" + \
                                "translation of layer " + poLayer.GetName() + " (use -skipfailures to skip errors)")
                        return False
                else:
                    # No matching target layer : just consumes the features

                    poFeature = poLayer.GetNextFeature()
                    while poFeature is not None:
                        anReadFeatureCount[0] = anReadFeatureCount[0] + 1
                        poFeature = poLayer.GetNextFeature()

                if anReadFeatureCount[0] != 0:
                    bHasLayersNonEmpty = True

    else:

        nLayerCount = 0
        papoLayers = []

# --------------------------------------------------------------------
#      Process each data source layer.
# --------------------------------------------------------------------
        if len(papszLayers) == 0:
            nLayerCount = poDS.GetLayerCount()
            papoLayers = [None for i in range(nLayerCount)]
            iLayer = 0

            for iLayer in range(nLayerCount):
                poLayer = poDS.GetLayer(iLayer)

                if poLayer is None:
                    print("FAILURE: Couldn't fetch advertised layer %d!" % iLayer)
                    return False

                papoLayers[iLayer] = poLayer
                iLayer = iLayer + 1

# --------------------------------------------------------------------
#      Process specified data source layers.
# --------------------------------------------------------------------
        else:
            nLayerCount = len(papszLayers)
            papoLayers = [None for i in range(nLayerCount)]
            iLayer = 0

            for layername in papszLayers:
                poLayer = poDS.GetLayerByName(layername)

                if poLayer is None:
                    print("FAILURE: Couldn't fetch advertised layer %s!" % layername)
                    return False

                papoLayers[iLayer] = poLayer
                iLayer = iLayer + 1

        panLayerCountFeatures = [0 for i in range(nLayerCount)]
        nCountLayersFeatures = 0
        nAccCountFeatures = 0

        # First pass to apply filters and count all features if necessary
        for iLayer in range(nLayerCount):
            poLayer = papoLayers[iLayer]

            if pszWHERE is not None:
                if poLayer.SetAttributeFilter( pszWHERE ) != 0:
                    print("FAILURE: SetAttributeFilter(%s) failed." % pszWHERE)
                    if not bSkipFailures:
                        return False

            if poSpatialFilter is not None:
                poLayer.SetSpatialFilter( poSpatialFilter )

            if bDisplayProgress and not bSrcIsOSM:
                if not poLayer.TestCapability(ogr.OLCFastFeatureCount):
                    print("Progress turned off as fast feature count is not available.")
                    bDisplayProgress = False
                else:
                    panLayerCountFeatures[iLayer] = poLayer.GetFeatureCount()
                    nCountLayersFeatures += panLayerCountFeatures[iLayer]

        # Second pass to do the real job
        for iLayer in range(nLayerCount):
            poLayer = papoLayers[iLayer]

            if bDisplayProgress:
                if bSrcIsOSM:
                    pfnProgress = progress_func
                    pProgressArg = progress_data
                else:
                    pfnProgress = ScaledProgressFunc
                    pProgressArg = ScaledProgressObject( \
                            nAccCountFeatures * 1.0 / nCountLayersFeatures, \
                            (nAccCountFeatures + panLayerCountFeatures[iLayer]) * 1.0 / nCountLayersFeatures, \
                            progress_func, progress_data)

            nAccCountFeatures += panLayerCountFeatures[iLayer]

# --------------------------------------------------------------------
#      Special case to improve user experience when translating into
#      single file shapefile and source has only one layer, and that
#      the layer name isn't specified
# --------------------------------------------------------------------
            if EQUAL(poDriver.GetName(), "ESRI Shapefile") and \
                nLayerCount == 1 and pszNewLayerName is None:
                try:
                    mode = os.stat(pszDestDataSource).st_mode
                    if (mode & stat.S_IFDIR) == 0:
                        pszNewLayerName = os.path.splitext(os.path.basename(pszDestDataSource))[0]
                except:
                    pass


            psInfo = SetupTargetLayer( poDS, \
                                       poLayer, \
                                       poODS, \
                                       papszLCO, \
                                       pszNewLayerName, \
                                       bTransform, \
                                       poOutputSRS, \
                                       bNullifyOutputSRS, \
                                       poSourceSRS, \
                                       papszSelFields, \
                                       bAppend, eGType, bPromoteToMulti, nCoordDim, bOverwrite, \
                                       papszFieldTypesToString, \
                                       bWrapDateline, \
                                       bExplodeCollections, \
                                       pszZField, \
                                       pszWHERE )

            poLayer.ResetReading()

            if (psInfo is None or \
                not TranslateLayer( psInfo, poDS, poLayer, poODS, \
                                    poOutputSRS, bNullifyOutputSRS, \
                                    eGType, bPromoteToMulti, nCoordDim, \
                                    eGeomOp, dfGeomOpParam, \
                                    panLayerCountFeatures[iLayer], \
                                    poClipSrc, poClipDst, \
                                    bExplodeCollections, \
                                    nSrcFileSize, None, \
                                    pfnProgress, pProgressArg )) \
                and not bSkipFailures:
                '''
                print(
                        "Terminating translation prematurely after failed\n" + \
                        "translation of layer " + poLayer.GetLayerDefn().GetName() + " (use -skipfailures to skip errors)")
                '''
                return False

# --------------------------------------------------------------------
#      Close down.
# --------------------------------------------------------------------
    # We must explicitly destroy the output dataset in order the file
    # to be properly closed !
    poODS.Destroy()
    poDS.Destroy()

    return True

#**********************************************************************
#                               Usage()
#**********************************************************************

def Usage():

    print( "Usage: ogr2ogr [--help-general] [-skipfailures] [-append] [-update] [-gt n]\n" + \
            "               [-select field_list] [-where restricted_where] \n" + \
            "               [-progress] [-sql <sql statement>] \n" + \
            "               [-spat xmin ymin xmax ymax] [-preserve_fid] [-fid FID]\n" + \
            "               [-a_srs srs_def] [-t_srs srs_def] [-s_srs srs_def]\n" + \
            "               [-f format_name] [-overwrite] [[-dsco NAME=VALUE] ...]\n" + \
            "               [-simplify tolerance]\n" + \
            #// "               [-segmentize max_dist] [-fieldTypeToString All|(type1[,type2]*)]\n" + \
            "               [-fieldTypeToString All|(type1[,type2]*)] [-explodecollections] \n" + \
            "               dst_datasource_name src_datasource_name\n" + \
            "               [-lco NAME=VALUE] [-nln name] [-nlt type] [-dim 2|3] [layer [layer ...]]\n" + \
            "\n" + \
            " -f format_name: output file format name, possible values are:")

    for iDriver in range(ogr.GetDriverCount()):
        poDriver = ogr.GetDriver(iDriver)

        if poDriver.TestCapability( ogr.ODrCCreateDataSource ):
            print( "     -f \"" + poDriver.GetName() + "\"" )

    print( " -append: Append to existing layer instead of creating new if it exists\n" + \
            " -overwrite: delete the output layer and recreate it empty\n" + \
            " -update: Open existing output datasource in update mode\n" + \
            " -progress: Display progress on terminal. Only works if input layers have the \"fast feature count\" capability\n" + \
            " -select field_list: Comma-delimited list of fields from input layer to\n" + \
            "                     copy to the new layer (defaults to all)\n" + \
            " -where restricted_where: Attribute query (like SQL WHERE)\n" + \
            " -sql statement: Execute given SQL statement and save result.\n" + \
            " -skipfailures: skip features or layers that fail to convert\n" + \
            " -gt n: group n features per transaction (default 200)\n" + \
            " -spat xmin ymin xmax ymax: spatial query extents\n" + \
            " -simplify tolerance: distance tolerance for simplification.\n" + \
            #//" -segmentize max_dist: maximum distance between 2 nodes.\n" + \
            #//"                       Used to create intermediate points\n" + \
            " -dsco NAME=VALUE: Dataset creation option (format specific)\n" + \
            " -lco  NAME=VALUE: Layer creation option (format specific)\n" + \
            " -nln name: Assign an alternate name to the new layer\n" + \
            " -nlt type: Force a geometry type for new layer.  One of NONE, GEOMETRY,\n" + \
            "      POINT, LINESTRING, POLYGON, GEOMETRYCOLLECTION, MULTIPOINT,\n" + \
            "      MULTIPOLYGON, or MULTILINESTRING.  Add \"25D\" for 3D layers.\n" + \
            "      Default is type of source layer.\n" + \
            " -dim dimension: Force the coordinate dimension to the specified value.\n" + \
            " -fieldTypeToString type1,...: Converts fields of specified types to\n" + \
            "      fields of type string in the new layer. Valid types are : \n" + \
            "      Integer, Real, String, Date, Time, DateTime, Binary, IntegerList, RealList,\n" + \
            "      StringList. Special value All can be used to convert all fields to strings.")

    print(" -a_srs srs_def: Assign an output SRS\n"
        " -t_srs srs_def: Reproject/transform to this SRS on output\n"
        " -s_srs srs_def: Override source SRS\n"
        "\n"
        " Srs_def can be a full WKT definition (hard to escape properly),\n"
        " or a well known definition (i.e. EPSG:4326) or a file with a WKT\n"
        " definition." )

    return False

def CSLFindString(v, mystr):
    i = 0
    for strIter in v:
        if EQUAL(strIter, mystr):
            return i
        i = i + 1
    return -1

def IsNumber( pszStr):
    try:
        (float)(pszStr)
        return True
    except:
        return False

def LoadGeometry( pszDS, pszSQL, pszLyr, pszWhere):
    poGeom = None

    poDS = ogr.Open( pszDS, False )
    if poDS is None:
        return None

    if pszSQL is not None:
        poLyr = poDS.ExecuteSQL( pszSQL, None, None )
    elif pszLyr is not None:
        poLyr = poDS.GetLayerByName(pszLyr)
    else:
        poLyr = poDS.GetLayer(0)

    if poLyr is None:
        print("Failed to identify source layer from datasource.")
        poDS.Destroy()
        return None

    if pszWhere is not None:
        poLyr.SetAttributeFilter(pszWhere)

    poFeat = poLyr.GetNextFeature()
    while poFeat is not None:
        poSrcGeom = poFeat.GetGeometryRef()
        if poSrcGeom is not None:
            eType = wkbFlatten(poSrcGeom.GetGeometryType())

            if poGeom is None:
                poGeom = ogr.Geometry( ogr.wkbMultiPolygon )

            if eType == ogr.wkbPolygon:
                poGeom.AddGeometry( poSrcGeom )
            elif eType == ogr.wkbMultiPolygon:
                for iGeom in range(poSrcGeom.GetGeometryCount()):
                    poGeom.AddGeometry(poSrcGeom.GetGeometryRef(iGeom) )

            else:
                print("ERROR: Geometry not of polygon type." )
                if pszSQL is not None:
                    poDS.ReleaseResultSet( poLyr )
                poDS.Destroy()
                return None

        poFeat = poLyr.GetNextFeature()

    if pszSQL is not None:
        poDS.ReleaseResultSet( poLyr )
    poDS.Destroy()

    return poGeom


def wkbFlatten(x):
    return x & (~ogr.wkb25DBit)

#**********************************************************************
#                               SetZ()
#**********************************************************************

def SetZ (poGeom, dfZ ):

    if poGeom is None:
        return

    eGType = wkbFlatten(poGeom.GetGeometryType())
    if eGType == ogr.wkbPoint:
        poGeom.SetPoint(0, poGeom.GetX(), poGeom.GetY(), dfZ)

    elif eGType == ogr.wkbLineString or \
         eGType == ogr.wkbLinearRing:
        for i in range(poGeom.GetPointCount()):
            poGeom.SetPoint(i, poGeom.GetX(i), poGeom.GetY(i), dfZ)

    elif eGType == ogr.wkbPolygon or \
         eGType == ogr.wkbMultiPoint or \
         eGType == ogr.wkbMultiLineString or \
         eGType == ogr.wkbMultiPolygon or \
         eGType == ogr.wkbGeometryCollection:
        for i in range(poGeom.GetGeometryCount()):
            SetZ(poGeom.GetGeometryRef(i), dfZ)

#**********************************************************************
#                         SetupTargetLayer()
#**********************************************************************

def SetupTargetLayer( poSrcDS, poSrcLayer, poDstDS, papszLCO, pszNewLayerName, \
                    bTransform,  poOutputSRS, bNullifyOutputSRS, poSourceSRS, papszSelFields, \
                    bAppend, eGType, bPromoteToMulti, nCoordDim, bOverwrite, \
                    papszFieldTypesToString, bWrapDateline, \
                    bExplodeCollections, pszZField, pszWHERE) :

    if pszNewLayerName is None:
        pszNewLayerName = poSrcLayer.GetLayerDefn().GetName()

# --------------------------------------------------------------------
#      Setup coordinate transformation if we need it.
# --------------------------------------------------------------------
    poCT = None

    if bTransform:
        if poSourceSRS is None:
            poSourceSRS = poSrcLayer.GetSpatialRef()

        if poSourceSRS is None:
            print("Can't transform coordinates, source layer has no\n" + \
                    "coordinate system.  Use -s_srs to set one." )
            return None

        poCT = osr.CoordinateTransformation( poSourceSRS, poOutputSRS )
        if gdal.GetLastErrorMsg().find( 'Unable to load PROJ.4 library' ) != -1:
            poCT = None

        if poCT is None:
            pszWKT = None

            print("Failed to create coordinate transformation between the\n" + \
                "following coordinate systems.  This may be because they\n" + \
                "are not transformable, or because projection services\n" + \
                "(PROJ.4 DLL/.so) could not be loaded." )

            pszWKT = poSourceSRS.ExportToPrettyWkt( 0 )
            print( "Source:\n" + pszWKT )

            pszWKT = poOutputSRS.ExportToPrettyWkt( 0 )
            print( "Target:\n" + pszWKT )
            return None

# --------------------------------------------------------------------
#      Get other info.
# --------------------------------------------------------------------
    poSrcFDefn = poSrcLayer.GetLayerDefn()

    if poOutputSRS is None and not bNullifyOutputSRS:
        poOutputSRS = poSrcLayer.GetSpatialRef()

# --------------------------------------------------------------------
#      Find the layer.
# --------------------------------------------------------------------

    # GetLayerByName() can instantiate layers that would have been
    # 'hidden' otherwise, for example, non-spatial tables in a
    # PostGIS-enabled database, so this apparently useless command is
    # not useless. (#4012)
    gdal.PushErrorHandler('CPLQuietErrorHandler')
    poDstLayer = poDstDS.GetLayerByName(pszNewLayerName)
    gdal.PopErrorHandler()
    gdal.ErrorReset()

    iLayer = -1
    if poDstLayer is not None:
        nLayerCount = poDstDS.GetLayerCount()
        for iLayer in range(nLayerCount):
            poLayer = poDstDS.GetLayer(iLayer)
            # The .cpp version compares on pointers directly, but we cannot
            # do this with swig object, so just compare the names.
            if poLayer is not None \
                and poLayer.GetName() == poDstLayer.GetName():
                break

        if (iLayer == nLayerCount):
            # Shouldn't happen with an ideal driver
            poDstLayer = None

# --------------------------------------------------------------------
#      If the user requested overwrite, and we have the layer in
#      question we need to delete it now so it will get recreated
#      (overwritten).
# --------------------------------------------------------------------
    if poDstLayer is not None and bOverwrite:
        if poDstDS.DeleteLayer( iLayer ) != 0:
            print("DeleteLayer() failed when overwrite requested." )
            return None

        poDstLayer = None

# --------------------------------------------------------------------
#      If the layer does not exist, then create it.
# --------------------------------------------------------------------
    if poDstLayer is None:
        if eGType == -2:
            eGType = poSrcFDefn.GetGeomType()

            n25DBit = eGType & ogr.wkb25DBit
            if bPromoteToMulti:
                if wkbFlatten(eGType) == ogr.wkbLineString:
                    eGType = ogr.wkbMultiLineString | n25DBit
                elif wkbFlatten(eGType) == ogr.wkbPolygon:
                    eGType = ogr.wkbMultiPolygon | n25DBit

            if bExplodeCollections:
                if wkbFlatten(eGType) == ogr.wkbMultiPoint:
                    eGType = ogr.wkbPoint | n25DBit
                elif wkbFlatten(eGType) == ogr.wkbMultiLineString:
                    eGType = ogr.wkbLineString | n25DBit
                elif wkbFlatten(eGType) == ogr.wkbMultiPolygon:
                    eGType = ogr.wkbPolygon | n25DBit
                elif wkbFlatten(eGType) == ogr.wkbGeometryCollection:
                    eGType = ogr.wkbUnknown | n25DBit

            if pszZField is not None:
                eGType = eGType | ogr.wkb25DBit

        if nCoordDim == 2:
            eGType = eGType & ~ogr.wkb25DBit
        elif nCoordDim == 3:
            eGType = eGType | ogr.wkb25DBit

        if poDstDS.TestCapability( ogr.ODsCCreateLayer ) == False:
            #print("Layer " + pszNewLayerName + "not found, and CreateLayer not supported by driver.")
            return None

        gdal.ErrorReset()

        poDstLayer = poDstDS.CreateLayer( pszNewLayerName, poOutputSRS, \
                                            eGType, papszLCO )

        if poDstLayer is None:
            return None

        bAppend = False

# --------------------------------------------------------------------
#      Otherwise we will append to it, if append was requested.
# --------------------------------------------------------------------
    elif not bAppend:
        print("FAILED: Layer " + pszNewLayerName + "already exists, and -append not specified.\n" + \
                            "        Consider using -append, or -overwrite.")
        return None
    else:
        if len(papszLCO) > 0:
            print("WARNING: Layer creation options ignored since an existing layer is\n" + \
                    "         being appended to." )

# --------------------------------------------------------------------
#      Add fields.  Default to copy all field.
#      If only a subset of all fields requested, then output only
#      the selected fields, and in the order that they were
#      selected.
# --------------------------------------------------------------------

    # Initialize the index-to-index map to -1's
    nSrcFieldCount = poSrcFDefn.GetFieldCount()
    panMap = [ -1 for i in range(nSrcFieldCount) ]

    poDstFDefn = poDstLayer.GetLayerDefn()

    if papszSelFields is not None and not bAppend:

        nDstFieldCount = 0
        if poDstFDefn is not None:
            nDstFieldCount = poDstFDefn.GetFieldCount()

        for iField in range(len(papszSelFields)):

            iSrcField = poSrcFDefn.GetFieldIndex(papszSelFields[iField])
            if iSrcField >= 0:
                poSrcFieldDefn = poSrcFDefn.GetFieldDefn(iSrcField)
                oFieldDefn = ogr.FieldDefn( poSrcFieldDefn.GetNameRef(),
                                            poSrcFieldDefn.GetType() )
                oFieldDefn.SetWidth( poSrcFieldDefn.GetWidth() )
                oFieldDefn.SetPrecision( poSrcFieldDefn.GetPrecision() )

                if papszFieldTypesToString is not None and \
                    (CSLFindString(papszFieldTypesToString, "All") != -1 or \
                    CSLFindString(papszFieldTypesToString, \
                                ogr.GetFieldTypeName(poSrcFieldDefn.GetType())) != -1):

                    oFieldDefn.SetType(ogr.OFTString)

                # The field may have been already created at layer creation
                iDstField = -1
                if poDstFDefn is not None:
                    iDstField = poDstFDefn.GetFieldIndex(oFieldDefn.GetNameRef())
                if iDstField >= 0:
                    panMap[iSrcField] = iDstField
                elif poDstLayer.CreateField( oFieldDefn ) == 0:
                    # now that we've created a field, GetLayerDefn() won't return NULL
                    if poDstFDefn is None:
                        poDstFDefn = poDstLayer.GetLayerDefn()

                    # Sanity check : if it fails, the driver is buggy
                    if poDstFDefn is not None and \
                        poDstFDefn.GetFieldCount() != nDstFieldCount + 1:
                        print("The output driver has claimed to have added the %s field, but it did not!" %  oFieldDefn.GetNameRef() )
                    else:
                        panMap[iSrcField] = nDstFieldCount
                        nDstFieldCount = nDstFieldCount + 1

            else:
                print("Field '" + papszSelFields[iField] + "' not found in source layer.")
                if not bSkipFailures:
                    return None

        # --------------------------------------------------------------------
        # Use SetIgnoredFields() on source layer if available
        # --------------------------------------------------------------------

        # Here we differ from the ogr2ogr.cpp implementation since the OGRFeatureQuery
        # isn't mapped to swig. So in that case just don't use SetIgnoredFields()
        # to avoid issue raised in #4015
        if poSrcLayer.TestCapability(ogr.OLCIgnoreFields) and pszWHERE is None:
            papszIgnoredFields = []
            for iSrcField in range(nSrcFieldCount):
                pszFieldName = poSrcFDefn.GetFieldDefn(iSrcField).GetNameRef()
                bFieldRequested = False
                for iField in range(len(papszSelFields)):
                    if EQUAL(pszFieldName, papszSelFields[iField]):
                        bFieldRequested = True
                        break

                if pszZField is not None and EQUAL(pszFieldName, pszZField):
                    bFieldRequested = True

                # If source field not requested, add it to ignored files list
                if not bFieldRequested:
                    papszIgnoredFields.append(pszFieldName)

            poSrcLayer.SetIgnoredFields(papszIgnoredFields)

    elif not bAppend:

        nDstFieldCount = 0
        if poDstFDefn is not None:
            nDstFieldCount = poDstFDefn.GetFieldCount()

        for iField in range(nSrcFieldCount):

            poSrcFieldDefn = poSrcFDefn.GetFieldDefn(iField)
            oFieldDefn = ogr.FieldDefn( poSrcFieldDefn.GetNameRef(),
                                        poSrcFieldDefn.GetType() )
            oFieldDefn.SetWidth( poSrcFieldDefn.GetWidth() )
            oFieldDefn.SetPrecision( poSrcFieldDefn.GetPrecision() )

            if papszFieldTypesToString is not None and \
                (CSLFindString(papszFieldTypesToString, "All") != -1 or \
                CSLFindString(papszFieldTypesToString, \
                            ogr.GetFieldTypeName(poSrcFieldDefn.GetType())) != -1):

                oFieldDefn.SetType(ogr.OFTString)

            # The field may have been already created at layer creation
            iDstField = -1
            if poDstFDefn is not None:
                 iDstField = poDstFDefn.GetFieldIndex(oFieldDefn.GetNameRef())
            if iDstField >= 0:
                panMap[iField] = iDstField
            elif poDstLayer.CreateField( oFieldDefn ) == 0:
                # now that we've created a field, GetLayerDefn() won't return NULL
                if poDstFDefn is None:
                    poDstFDefn = poDstLayer.GetLayerDefn()

                # Sanity check : if it fails, the driver is buggy
                if poDstFDefn is not None and \
                    poDstFDefn.GetFieldCount() != nDstFieldCount + 1:
                    print("The output driver has claimed to have added the %s field, but it did not!" %  oFieldDefn.GetNameRef() )
                else:
                    panMap[iField] = nDstFieldCount
                    nDstFieldCount = nDstFieldCount + 1

    else:
        # For an existing layer, build the map by fetching the index in the destination
        # layer for each source field
        if poDstFDefn is None:
            print( "poDstFDefn == NULL.\n" )
            return None

        for iField in range(nSrcFieldCount):
            poSrcFieldDefn = poSrcFDefn.GetFieldDefn(iField)
            iDstField = poDstFDefn.GetFieldIndex(poSrcFieldDefn.GetNameRef())
            if iDstField >= 0:
                panMap[iField] = iDstField

    iSrcZField = -1
    if pszZField is not None:
        iSrcZField = poSrcFDefn.GetFieldIndex(pszZField)

    psInfo = TargetLayerInfo()
    psInfo.poDstLayer = poDstLayer
    psInfo.poCT = poCT
    #psInfo.papszTransformOptions = papszTransformOptions
    psInfo.panMap = panMap
    psInfo.iSrcZField = iSrcZField

    return psInfo

#**********************************************************************
#                           TranslateLayer()
#**********************************************************************

def TranslateLayer( psInfo, poSrcDS, poSrcLayer, poDstDS,  \
                    poOutputSRS, bNullifyOutputSRS, \
                    eGType, bPromoteToMulti, nCoordDim, eGeomOp, dfGeomOpParam, \
                    nCountLayerFeatures, \
                    poClipSrc, poClipDst, bExplodeCollections, nSrcFileSize, \
                    pnReadFeatureCount, pfnProgress, pProgressArg) :

    bForceToPolygon = False
    bForceToMultiPolygon = False
    bForceToMultiLineString = False

    poDstLayer = psInfo.poDstLayer
    #papszTransformOptions = psInfo.papszTransformOptions
    poCT = psInfo.poCT
    panMap = psInfo.panMap
    iSrcZField = psInfo.iSrcZField

    if poOutputSRS is None and not bNullifyOutputSRS:
        poOutputSRS = poSrcLayer.GetSpatialRef()

    if wkbFlatten(eGType) == ogr.wkbPolygon:
        bForceToPolygon = True
    elif wkbFlatten(eGType) == ogr.wkbMultiPolygon:
        bForceToMultiPolygon = True
    elif wkbFlatten(eGType) == ogr.wkbMultiLineString:
        bForceToMultiLineString = True

# --------------------------------------------------------------------
#      Transfer features.
# --------------------------------------------------------------------
    nFeaturesInTransaction = 0
    nCount = 0

    if nGroupTransactions > 0:
        poDstLayer.StartTransaction()

    while True:
        poDstFeature = None

        if nFIDToFetch != ogr.NullFID:

            #// Only fetch feature on first pass.
            if nFeaturesInTransaction == 0:
                poFeature = poSrcLayer.GetFeature(nFIDToFetch)
            else:
                poFeature = None

        else:
            poFeature = poSrcLayer.GetNextFeature()

        if poFeature is None:
            break

        nParts = 0
        nIters = 1
        if bExplodeCollections:
            poSrcGeometry = poFeature.GetGeometryRef()
            if poSrcGeometry is not None:
                eSrcType = wkbFlatten(poSrcGeometry.GetGeometryType())
                if eSrcType == ogr.wkbMultiPoint or \
                   eSrcType == ogr.wkbMultiLineString or \
                   eSrcType == ogr.wkbMultiPolygon or \
                   eSrcType == ogr.wkbGeometryCollection:
                        nParts = poSrcGeometry.GetGeometryCount()
                        nIters = nParts
                        if nIters == 0:
                            nIters = 1

        for iPart in range(nIters):
            nFeaturesInTransaction = nFeaturesInTransaction + 1
            if nFeaturesInTransaction == nGroupTransactions:
                poDstLayer.CommitTransaction()
                poDstLayer.StartTransaction()
                nFeaturesInTransaction = 0

            gdal.ErrorReset()
            poDstFeature = ogr.Feature( poDstLayer.GetLayerDefn() )

            if poDstFeature.SetFromWithMap( poFeature, 1, panMap ) != 0:

                if nGroupTransactions > 0:
                    poDstLayer.CommitTransaction()

                print("Unable to translate feature %d from layer %s" % (poFeature.GetFID() , poSrcLayer.GetName() ))

                return False

            if bPreserveFID:
                poDstFeature.SetFID( poFeature.GetFID() )

            poDstGeometry = poDstFeature.GetGeometryRef()
            if poDstGeometry is not None:

                if nParts > 0:
                    # For -explodecollections, extract the iPart(th) of the geometry
                    poPart = poDstGeometry.GetGeometryRef(iPart).Clone()
                    poDstFeature.SetGeometryDirectly(poPart)
                    poDstGeometry = poPart

                if iSrcZField != -1:
                    SetZ(poDstGeometry, poFeature.GetFieldAsDouble(iSrcZField))
                    # This will correct the coordinate dimension to 3
                    poDupGeometry = poDstGeometry.Clone()
                    poDstFeature.SetGeometryDirectly(poDupGeometry)
                    poDstGeometry = poDupGeometry


                if nCoordDim == 2 or nCoordDim == 3:
                    poDstGeometry.SetCoordinateDimension( nCoordDim )

                if eGeomOp == GeomOperation.SEGMENTIZE:
                    pass
                    #if (poDstFeature.GetGeometryRef() is not None and dfGeomOpParam > 0)
                    #    poDstFeature.GetGeometryRef().segmentize(dfGeomOpParam);
                elif eGeomOp == GeomOperation.SIMPLIFY_PRESERVE_TOPOLOGY and dfGeomOpParam > 0:
                    poNewGeom = poDstGeometry.SimplifyPreserveTopology(dfGeomOpParam)
                    if poNewGeom is not None:
                        poDstFeature.SetGeometryDirectly(poNewGeom)
                        poDstGeometry = poNewGeom

                if poClipSrc is not None:
                    poClipped = poDstGeometry.Intersection(poClipSrc)
                    if poClipped is None or poClipped.IsEmpty():
                        # Report progress
                        nCount = nCount +1
                        if pfnProgress is not None:
                            pfnProgress(nCount * 1.0 / nCountLayerFeatures, "", pProgressArg)
                        continue

                    poDstFeature.SetGeometryDirectly(poClipped)
                    poDstGeometry = poClipped

                if poCT is not None:
                    eErr = poDstGeometry.Transform( poCT )
                    if eErr != 0:
                        if nGroupTransactions > 0:
                            poDstLayer.CommitTransaction()

                        print("Failed to reproject feature %d (geometry probably out of source or destination SRS)." % poFeature.GetFID())
                        if not bSkipFailures:
                            return False

                elif poOutputSRS is not None:
                    poDstGeometry.AssignSpatialReference(poOutputSRS)

                if poClipDst is not None:
                    poClipped = poDstGeometry.Intersection(poClipDst)
                    if poClipped is None or poClipped.IsEmpty():
                        continue

                    poDstFeature.SetGeometryDirectly(poClipped)
                    poDstGeometry = poClipped

                if bForceToPolygon:
                    poDstFeature.SetGeometryDirectly(ogr.ForceToPolygon(poDstGeometry))

                elif bForceToMultiPolygon or \
                        (bPromoteToMulti and wkbFlatten(poDstGeometry.GetGeometryType()) == ogr.wkbPolygon):
                    poDstFeature.SetGeometryDirectly(ogr.ForceToMultiPolygon(poDstGeometry))

                elif bForceToMultiLineString or \
                        (bPromoteToMulti and wkbFlatten(poDstGeometry.GetGeometryType()) == ogr.wkbLineString):
                    poDstFeature.SetGeometryDirectly(ogr.ForceToMultiLineString(poDstGeometry))

            gdal.ErrorReset()
            if poDstLayer.CreateFeature( poDstFeature ) != 0 and not bSkipFailures:
                if nGroupTransactions > 0:
                    poDstLayer.RollbackTransaction()

                return False

        # Report progress
        nCount = nCount  + 1
        if pfnProgress is not None:
            if nSrcFileSize != 0:
                if (nCount % 1000) == 0:
                    poFCLayer = poSrcDS.ExecuteSQL("GetBytesRead()", None, None)
                    if poFCLayer is not None:
                        poFeat = poFCLayer.GetNextFeature()
                        if poFeat is not None:
                            pszReadSize = poFeat.GetFieldAsString(0)
                            nReadSize = int(pszReadSize)
                            pfnProgress(nReadSize * 1.0 / nSrcFileSize, "", pProgressArg)
                    poSrcDS.ReleaseResultSet(poFCLayer)
            else:
                pfnProgress(nCount * 1.0 / nCountLayerFeatures, "", pProgressArg)

        if pnReadFeatureCount is not None:
            pnReadFeatureCount[0] = nCount

    if nGroupTransactions > 0:
        poDstLayer.CommitTransaction()

    return True

if __name__ == '__main__':
    version_num = int(gdal.VersionInfo('VERSION_NUM'))
    if version_num < 1800: # because of ogr.GetFieldTypeName
        print('ERROR: Python bindings of GDAL 1.8.0 or later required')
        sys.exit(1)

    if not main(sys.argv):
        sys.exit(1)
    else:
        sys.exit(0)



'''
@author: Niklas Aßelmann
'''

import xml.etree.ElementTree as ET  
import helpfunctions as hf
import ogr2ogr
import sys, os
from osgeo import gdal, ogr
import convex_hull


DATATYPE = "application/xml"

def isValid(filePath):
    '''Checks whether it is valid XML or not. \n
    input "path": type string, path to file which shall be extracted \n
    output true if file is valid, false if not
    '''
    try:
        with open(filePath) as XML_file:
            tree = ET.parse(XML_file)
            root = tree.getroot()
            if root is None:
                raise Exception('The xml file from ' + filePath + ' has no valid xml Attributes')
            else:
                return True
    except:
        raise Exception('The xml file from ' + filePath + ' has no valid xml Attributes')

def getBoundingBox(filePath):
    '''
    extract bounding box from xml \n
    input "filepath": type string, file path to xml file \n
    returns bounding box of the file: type list, length = 4 , type = float, schema = [min(longs), min(lats), max(longs), max(lats)]
    '''
    with open(filePath) as XML_file:
        lat = []
        lon = []
        tree = ET.parse(XML_file)
        root = tree.getroot()
        for x in root:
            if x.find('lat') is not None:  
                latitute = x.find('lat').text
                lat.append(latitute)
            else:
                if x.find('latitude') is not None:
                    latitute = x.find('latitude').text
                    lat.append(latitute)
            if x.find('lon') is not None:  
                longitude = x.find('lon').text
                lon.append(longitude)
            else:
                if x.find('longitude') is not None:
                    longitude = x.find('longitude').text
                    lon.append(longitude)
        spatialExtent={}
        if lat is not None:
            minlat=min(lat)
            maxlat=max(lat)
            if lon is not None:
                minlon=min(lon)
                maxlon=max(lon)
                spatialExtent= [minlon,minlat,maxlon,maxlat]
                return spatialExtent
            else:
                raise Exception('The xml file from ' + filePath + ' has no BoundingBox')
        else:
            raise Exception('The xml file from ' + filePath + ' has no BoundingBox')





def getTemporalExtent(filePath):
    '''
    extracts temporal extent of the xml \n
    input "filepath": type string, file path to xml file \n
    returns temporal extent: type list, length = 2, both entries have the type dateTime, temporalExtent[0] <= temporalExtent[1]
    '''
    with open(filePath) as XML_File:
        tree = ET.parse(XML_File)
        root = tree.getroot()
        alltime = []
        for x in root:
            if x.find('time') is not None:  
                time = x.find('time').text
                alltime.append(time)
        minTime= None
        maxTime= None
        if alltime is not None:
            minTime=min(alltime)
            maxTime=max(alltime)
            time=[]
            time.append(minTime)
            time.append(maxTime)
            return time
        else:
            raise Exception('The xml file from ' + filePath + ' has no TemporalExtent')




def getVectorRepresentation(filePath):
    '''
    extracts coordinates from xml File (for vector representation) \n
    input "filepath": type string, file path to xml file \n
    returns extracted coordinates of content from xml file: type list, list of lists with length = 2
    '''
    with open(filePath) as XML_file:
        lat = []
        lon = []
        tree = ET.parse(XML_file)
        root = tree.getroot()
        for x in root:
            if x.find('lat') is not None:  
                latitute = x.find('lat').text
                lat.append(latitute)
            else:
                if x.find('latitude') is not None:
                    latitute = x.find('latitude').text
                    lat.append(latitute)
            if x.find('lon') is not None:  
                longitude = x.find('lon').text
                lon.append(longitude)
            else:
                if x.find('longitude') is not None:
                    longitude = x.find('longitude').text
                    lon.append(longitude)
        vectorArray=[]
        if lon is None:
            raise Exception('The xml file from ' + filePath + ' has no VectorRepresentation')
        else:
            if lat is None:
                raise Exception('The xml file from ' + filePath + ' has no VectorRepresentation')
            else:
                counter=0
                for x in lon:
                    singleArray=[]
                    singleArray.append(float(lon[counter]))
                    singleArray.append(float(lat[counter]))
                    vectorArray.append(singleArray)
                    counter=counter+1
            vectorArray= convex_hull.graham_scan(vectorArray)
            return vectorArray




def getCRS(filePath):
    '''
    extracts coordinatesystem from xml File \n
    input "filepath": type string, file path to xml file \n
    returns epsg code of the used coordiante references system: type list, contains extracted coordinate system of content from xml file
    '''
    with open(filePath) as XML_File:
        tree = ET.parse(XML_File)
        root = tree.getroot()
        coordinatesystem = []
        for x in root:
            if x.find('crs') is not None:  
                crs = x.find('crs').text
                coordinatesystem.append(crs)
        if coordinatesystem is None:
            raise Exception('The XML file from ' + filePath + ' has no CRS')
        if hf.searchForParameters(["crs","srsID"],coordinatesystem) == "WGS84" or "4326":
            return "4326"
        else:
            raise Exception('The XML file from ' + filePath + ' has no WGS84 CRS')



'''
@author: Katharina Hovestadt
'''
from osgeo import ogr
from datetime import datetime as dtime

def createXmlTree(metadata, uuid):
    '''
    input "metadata": type list, contains values for metadata that shall be updated
    input "uuid": type string, id from entry that shall be updated
    returns a xml string that is send as a post request to the server to update the entry in the database
    '''

    tempextent = ""
    vectorrep = ""
    bbox = ""
    #creating xml parts
    for key, value in metadata.items():
        if value:
            if key == "temporal_extent":
                if len(metadata["temporal_extent"])>1:
                    try:

                        startPoint = dtime.strptime(metadata["temporal_extent"][0] ,'%Y-%m-%dT%H:%M:%SZ')
                        endPoint = dtime.strptime(metadata["temporal_extent"][1] ,'%Y-%m-%dT%H:%M:%SZ')
                    
                        tempextent = '''<csw:RecordProperty>
                                        <csw:Name>apiso:TempExtent_begin</csw:Name>
                                        <csw:Value>''' + startPoint + '''</csw:Value>
                                    </csw:RecordProperty>
                                    <csw:RecordProperty>
                                        <csw:Name>apiso:TempExtent_end</csw:Name>
                                        <csw:Value>''' + endPoint + '''</csw:Value>
                                    </csw:RecordProperty>'''
                    except:
                        tempextent = ""
            if key == "vector_rep":
                if metadata["vector_rep"]:
                    #create wkt polygon by parsing coordinates to wkb 
                    wkbLineVecRep = ogr.Geometry(ogr.wkbLinearRing)
                    for elem in metadata["vector_rep"]:
                        wkbLineVecRep.AddPoint(elem[0], elem[1])
                    wkbLineVecRep.AddPoint(metadata["vector_rep"][0][0], metadata["vector_rep"][0][1])

                    wkbPolyVecRep = ogr.Geometry(ogr.wkbPolygon)
                    wkbPolyVecRep.AddGeometry(wkbLineVecRep)

                    wktVecRep = wkbPolyVecRep.ExportToWkt()

                    vectorrep = '''<csw:RecordProperty>
                                        <csw:Name>apiso:Vector_rep</csw:Name>
                                        <csw:Value>''' + wktVecRep + '''</csw:Value>
                                    </csw:RecordProperty>'''

            if key == "bbox":
                if metadata["bbox"]:
                    wkbLineBbox = ogr.Geometry(ogr.wkbLinearRing)
                    wkbLineBbox.AddPoint(metadata["bbox"][0],metadata["bbox"][1])
                    wkbLineBbox.AddPoint(metadata["bbox"][2],metadata["bbox"][3])
                    
                    wkbPolyBbox = ogr.Geometry(ogr.wkbPolygon)
                    wkbPolyBbox.AddGeometry(wkbLineBbox)

                    wktBbox = wkbPolyBbox.ExportToWkt()

                    bbox = '''<csw:RecordProperty>
                                <csw:Name>apiso:BoundingBox</csw:Name>
                                <csw:Value>''' + wktBbox + '''</csw:Value>
                            </csw:RecordProperty>'''

    # if there is nothing to update                   
    if tempextent == "" and vectorrep == "" and bbox == "":
        raise Exception("The given entry has no temporal extent, vector representation and bounding box. The file could not be updated.")

    
    return '''<?xml version="1.0" encoding="UTF-8"?>
        <csw:Transaction xmlns:ogc="http://www.opengis.net/ogc" xmlns:csw="http://www.opengis.net/cat/csw/2.0.2" xmlns:ows="http://www.opengis.net/ows" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.opengis.net/cat/csw/2.0.2 http://schemas.opengis.net/csw/2.0.2/CSW-publication.xsd" service="CSW" version="2.0.2">
            <csw:Update>''' + tempextent + vectorrep + bbox + '''<csw:Constraint version="1.1.0">
                    <ogc:Filter>
                        <ogc:PropertyIsEqualTo>
                            <ogc:PropertyName>apiso:Identifier</ogc:PropertyName>
                        <ogc:Literal>''' + uuid + '''</ogc:Literal>
                        </ogc:PropertyIsEqualTo>
                    </ogc:Filter>
                </csw:Constraint>
            </csw:Update>
        </csw:Transaction>
        '''



'''
@author: Katharina Hovestadt, Niklas Asselmann, Benjamin Dietz
'''
import sys, os, getopt, datetime, errno, sqlite3, subprocess, uuid # important
from six.moves import configparser
from os import walk
import helpfunctions as hf
import threading 
import convex_hull

def computeBboxInWGS84(module, path):
    ''' input "module": type module, module from which methods shall be used \n
    input "path": type string, path to file \n
    returns a bounding box, type list, length = 4 , type = float, schema = [min(longs), min(lats), max(longs), max(lats)], the boudning box has either its original crs or WGS84 (transformed).
    '''
    bbox_in_orig_crs = module.getBoundingBox(path)
    try:
        crs = module.getCRS(path)
    except:
        pass
    if 'crs' in locals() and crs and bbox_in_orig_crs:
        bbox_transformed = hf.transformingArrayIntoWGS84(crs, bbox_in_orig_crs)
        return bbox_transformed
    else:
        raise Exception("The bounding box could not be related to a CRS")

def computeVectorRepresentationInWGS84(module, path):
    ''' input "module": type module, module from which methods shall be used \n
    input "path": type string, path to file \n
    returns a vector representation, type list, schema = [[lon1, lon2], [lon2, lon2], ...], that is in its original crs.
    '''
    crs = None
    vector_rep_in_orig_crs = module.getVectorRepresentation(path)
    try:
        crs = module.getCRS(path)
    except Exception as e:
        print("Exception in module.getCRS(path) in computeVectorRepresentationInWGS84(%s, %s): %s" % (module, path, e))

    if crs and vector_rep_in_orig_crs:
        vector_rep_transformed = hf.transformingArrayIntoWGS84(crs, vector_rep_in_orig_crs)
        return vector_rep_transformed
    else:
        raise Exception("The vector representation could not be related to a CRS.")




def extractMetadataFromFile(filePath, whatMetadata):
    ''' function is called when filePath is included in commanline (with tag 'e', 't' or 's')
    how this is done depends on the file format - the function calls the extractMetadataFrom<format>()-function \n
    input "filePath": type string, path to file from which the metadata shall be extracted \n
    input "whatMetadata": type string, specifices which metadata should be extracted  \n
    returns None if the format is not supported, else returns the metadata of the file as a dict 
    (possible) keys of the dict: 'temporal_extent', 'bbox', 'vector_reps', 'crs'
    '''
    
    fileFormat = filePath[filePath.rfind('.')+1:]
    usedModule = None

    # initialization of later output dict
    metadata = {}

    # first get the module that will be called (depending on the format of the file)
    if fileFormat == 'shp' or fileFormat == 'dbf':
        import handleShapefile
        usedModule = handleShapefile
    elif fileFormat == 'csv':
        import handleCSV
        usedModule = handleCSV
    elif fileFormat == 'nc':
        import handleNetCDF
        usedModule = handleNetCDF
    elif fileFormat == 'geojson' or fileFormat == 'json':
        import handleGeojson
        usedModule = handleGeojson
    elif fileFormat == 'gpkg':
        import handleGeopackage
        usedModule = handleGeopackage
    elif fileFormat == 'geotiff' or fileFormat == 'tif':
        import handleGeotiff
        usedModule = handleGeotiff
    elif fileFormat == 'gml':
        import handleGML
        usedModule = handleGML
    elif fileFormat == 'xml':
        import handleXML
        usedModule = handleXML
    elif fileFormat == 'kml':
        import handleKML
        usedModule = handleKML
    else: 
        # file format is not supported
        return None
    #only extracts metadata if the file content is valid
    try:
        valid = usedModule.isValid(filePath)
    except Exception as e:
        print("Error for " + filePath + ": " + str(e))
        valid = False 
    #get Bbox, Temporal Extent, Vector representation and crs parallel with threads
    class thread(threading.Thread): 
        def __init__(self, thread_ID): 
            threading.Thread.__init__(self) 
            self.thread_ID = thread_ID
        def run(self):
            metadata["format"] = usedModule.DATATYPE
            print("Thread with Thread_ID " +  str(self.thread_ID) + " now running...")
            #metadata[self.thread_ID] = self.thread_ID
            if self.thread_ID == 100:
                try:
                    metadata["bbox"] = computeBboxInWGS84(usedModule, filePath)
                except Exception as e:
                    print("Warning for " + filePath + ": " + str(e)) 
            elif self.thread_ID == 101:
                try:
                    metadata["temporal_extent"] = usedModule.getTemporalExtent(filePath)
                except Exception as e:
                    print("Warning for " + filePath + ": " + str(e))
            elif self.thread_ID == 102:
                try:
                    metadata["vector_rep"] = computeVectorRepresentationInWGS84(usedModule, filePath)
                except Exception as e:
                    print("Warning for " + filePath + ": " + str(e))
            
            elif self.thread_ID == 200:
                metadata["bbox"] = computeBboxInWGS84(usedModule, filePath)
            elif self.thread_ID == 201:
                metadata["temporal_extent"] = usedModule.getTemporalExtent(filePath)
            elif self.thread_ID == 202:
                metadata["vector_rep"] = computeVectorRepresentationInWGS84(usedModule, filePath)
            
            elif self.thread_ID == 103:
                try:
                    # the CRS is not neccessarily required
                    if hasattr(usedModule, 'getCRS'):
                        metadata["crs"] = usedModule.getCRS(filePath)
                    else: print ("Warning: The CRS cannot be extracted from the file")
                except Exception as e:
                    print("Warning for " + filePath + ": " + str(e))      
            try:
                barrier.wait()
            except Exception as e:
                barrier.abort()

            
    #thread id 100+ -> metadata extraction with exceptions from methods (raise Exception)
    #thread id 200+ -> metadata extraction without exceptions from methods ( only standard exceptions are raised (e.g. ValueError, AttributeError))
    thread_bbox_except = thread(100) 
    thread_temp_except = thread(101) 
    thread_vector_except = thread(102)
    thread_crs_except = thread(103)
    thread_bbox = thread(200) 
    thread_temp = thread(201) 
    thread_vector = thread(202) 
    
    if valid:
        if whatMetadata == "e":
            # none of the metadata field is required 
            # so the system does not crash even if it does not find anything
            barrier = threading.Barrier(5)
            thread_bbox_except.start() 
            thread_temp_except.start() 
            thread_vector_except.start() 
            thread_crs_except.start()
            barrier.wait() 
            barrier.reset() 
            barrier.abort() 
            
        if whatMetadata == "s":
            # only spatial extent is required
            # if one of the metadata field could not be extrated, the system crashes
            barrier = threading.Barrier(4)
            thread_bbox.start()
            thread_vector.start()
            thread_crs_except.start()
            barrier.wait()
            barrier.reset() 
            barrier.abort() 

        if whatMetadata == "t":
            # only temporal extent is required
            # if one of the metadata field could not be extracted, the system crashes
            barrier = threading.Barrier(2)
            thread_temp.start()
            barrier.wait()
            barrier.reset() 
            barrier.abort() 
    else:
                raise Exception("The file " + str(filePath) + " could not be validated")
        
    return metadata




def extractMetadataFromFolder(folderPath, whatMetadata):
    '''function is called when path of directory is included in commanline (with tag 'e', 't' or 's') \n
    calls the extractMetadataFromFile-function and computes the average for the metadata fields
    (possible) keys of the returning dict: 'temporal_extent', 'bbox', 'vector_reps' \n
    input "folderPath": type string, path to folder from which the metadata shall be extracted \n
    input "whatMetadata": type string, specifices which metadata should be extracted  \n
    returns the metadata of the folder as a dict \n
    '''
    if not os.path.isdir(folderPath):
        raise FileNotFoundError("This directory could not be found")
    else:
        # initialization of later output dict
        metadata = {}

        files_in_folder = []

        # get all files from the folder
        for (dirpath, dirnames, filenames) in walk(folderPath):
            files_in_folder.extend(filenames)
        # for shapefile: ignore all .prj-, .dbf-, .shx- and cpg-files where a related .shp-file exists
        # so that the same shapefile is only executed once
        i = 0
        while (i < len(files_in_folder)):
            pathWithoutEnding = files_in_folder[i][:len(files_in_folder[i])-4]
            if '.cpg' in files_in_folder[i] or '.prj' in files_in_folder[i] \
                or '.shx' in files_in_folder[i] or '.dbf' in files_in_folder[i]:
                if pathWithoutEnding + ".shp" in files_in_folder:
                    files_in_folder.remove(files_in_folder[i])
                else:
                    i += 1
            else:
                i += 1
        metadataElements = []
        fullPaths = []

        for x in files_in_folder:
            fullPaths.append(str(os.path.join(folderPath, x)))
        
        filesSkiped = 0
        bboxes = []
        temporal_extents = []
        vector_reps = []

        # handle each of the files in the folder seperately
        # get metadata fields 'bbox', 'vector_rep', 'temporal_extent' of all supported files
        for x in fullPaths:
            try:
                metadataOfFile = extractMetadataFromFile(x, "e")
                if metadataOfFile is not None:
                    metadataElements.append(metadataOfFile)
                    if 'bbox' in metadataOfFile:
                        if metadataOfFile["bbox"] is not None:
                            bboxes.append(metadataOfFile["bbox"])
                    if 'vector_rep' in metadataOfFile:
                        if metadataOfFile["vector_rep"] is not None:
                            vector_reps.append(metadataOfFile["vector_rep"]) # TO DO: here all the coors should be appended later
                    if 'temporal_extent' in metadataOfFile:
                        if metadataOfFile["temporal_extent"] is not None:
                            temporal_extents.append(metadataOfFile["temporal_extent"]) 
                else:
                    # fileformat is not supported
                    filesSkiped = filesSkiped + 1            
            except Exception as e:
                print("Warning for " + str(x) + ": "+ str(e))
    def getTemporalExtentFromFolder(mult_temp_extents):
        ''' computes temporal extent from multiple temporal extents stored in the array 'mult_temp_extents' (uses helpfunction) \n
        input "mult_temp_extents": type list, list of list with temporal extent with length = 2, both entries have the type dateTime, temporalExtent[0] <= temporalExtent[1] \n
        returns temporal extent of all files, type list with two datetime
        '''
        print(str(len(mult_temp_extents)) + " of " + str(len(fullPaths)-filesSkiped) + " supported Files have a temporal extent.")

        if len(mult_temp_extents) > 0 and hf.computeTempExtentOfMultiple(mult_temp_extents) is not None:
            return hf.computeTempExtentOfMultiple(mult_temp_extents)

        else: return None

    def getBboxFromFolder(mult_bboxes):
        ''' computes boundingbox from multiple bounding boxes stored in the array 'mult_bboxes' (uses helpfunction) \n
        input "mult_bboxes": type list, list with bounding boxes, one bbox has the following format:  length = 4 , type = float, schema = [min(longs), min(lats), max(longs), max(lats)] \n
        returns bounding box of the files in the folder, type list, length = 4 , type = float, schema = [min(longs), min(lats), max(longs), max(lats)], the boudning box has either its original crs or WGS84 (transformed).
        '''
        print(str(len(mult_bboxes)) + " of " + str(len(fullPaths)-filesSkiped) + " supported Files have a bbox.")
        
        if len(mult_bboxes) > 0 and hf.computeBboxOfMultiple(mult_bboxes) is not None:
            return hf.computeBboxOfMultiple(mult_bboxes)

        else: return None

    def getVectorRepFromFolder(mult_vec_rep):
        '''
        computes vector representation from multiple vector representations stored in the array 'mult_vec_rep' (uses helpfunction) \n
        input "mult_vec_rep": type list, all vector representations from the files in the folder \n
        returns the vector representation of the files in the folder: type list, one vector representation of the files from folder
        '''
        print("There are " + str(len(mult_vec_rep)) + " extracted coordinate pairs out of " + str(len(fullPaths)-filesSkiped) + " file(s).")
        if type(mult_vec_rep) == list:
            if len(mult_vec_rep) > 0:
                return convex_hull.graham_scan(mult_vec_rep)

        return None

    bbox = getBboxFromFolder(bboxes)
    if(type(vector_reps) == list):
        if(len(vector_reps) is not 0):    
            if type(vector_reps[0] == list):
                if type(vector_reps[0][0] == list):
                    vector_reps_help = []
                    for elem1 in vector_reps:
                        for elem2 in elem1:
                            vector_reps_help.append(elem2)
                    vector_reps = vector_reps_help
    vector_rep = getVectorRepFromFolder(vector_reps)
    temp_ext = getTemporalExtentFromFolder(temporal_extents)

    if whatMetadata == "e":
        
        if bbox is not None:
            metadata["bbox"] = bbox
        
        if vector_rep is not None:
            metadata["vector_rep"] = vector_rep
        
        if temp_ext is not None:
            metadata["temporal_extent"] = temp_ext

    
    if whatMetadata == "s":
        # raise exception if one of the metadata fields 'bbox' or 'vector_represenation' could not be extracted from the folder    
        if bbox is not None and vector_rep is not None:
            metadata["bbox"] = bbox
            metadata["vector_rep"] = vector_rep

        else: raise Exception("A spatial extent cannot be computed out of any file in this folder")

    if whatMetadata == "t":
        # raise exception if the temporal extent could not be extracted from the folder
        if temp_ext is not None:
            metadata["temporal_extent"] = temp_ext

        else: raise Exception("A temporal extent cannot be computed out of any file in this folder")
        
    if filesSkiped != 0: 
        print(str(filesSkiped) + ' file(s) has been skipped as its format is not supported; to see the supported formats look at -help')
    if len(fullPaths) - filesSkiped == 0:
        raise Exception("None of the files in the dictionary are supported.")
    
    return metadata



import pytest
import os
import os.path
import sqlite3
import filecmp  
import math
import SimScore


dasEntry1 = {"id" : 1,"wkt_geometry" : [45,50,30,45],"vector" : [[1,1], [0,9]],"time" : ["2000-07-16T19:20:30+01:00", "2001-07-16T19:20:30+01:00"],"raster" : False}
dasEntry2 = {"id" : 2,"wkt_geometry" : [50,55,10,12],"vector" : [[1,1],[2,12]],"time" : ["1999-07-16T19:20:30+01:00", "2000-07-16T19:20:30+01:00"],"raster" : True}
dasEntry3 = {"id" : 3,"wkt_geometry" : [20,25,4,8],"vector" : [[10,50],[3,54]],"time" : ["1998-07-16T19:20:30+01:00", "1999-07-16T19:20:30+01:00"],"raster" : True}
dasEntry4 = {"id" : 4,"wkt_geometry" : [20,22,23,28],"vector" : [[2,1],[3,5],[4,20]],"time" : ["1997-07-16T19:20:30+01:00", "1998-07-16T19:20:30+01:00"],"raster" : False}
dasDict = [dasEntry1,dasEntry2,dasEntry3,dasEntry4]

'''
 Function purpose: tests whether the correct results are put out
 when comparing two data types of geographic information
 input: -
 output:    1: same type and both vector or raster
            0.8: unequal type and both vector or raster
            0: unequal type and only a vector or raster
'''
def test_getGeoDatSim():
    entry1 = dasEntry1
    entry2 = dasEntry3
    geoSimilarity = SimScore.getGeoDatSim(entry1, entry2)
    print (geoSimilarity)
    # geoSimilarity = 0 || geoSimilarity = 0.8 || geoSimilarity = 1
    ans = 0
    if (entry1["raster"] == entry2["raster"]): #beides true, wenn es Raster ist. beides false, wenn es Vektor ist.
        if ((len(entry1["vector"]) == len(entry2["vector"])) or (len(entry1["vector"]) > 2 and len(entry2["vector"] > 2))): #wegen Punkt = 1, Linie = 2, Polygon > 2
            ans = 1
        else:
            ans = 0.8
    assert ans == geoSimilarity

'''
 Function purpose: Tests whether the correct results are output when comparing two types of temporal information
 input: arrays, entry1, entry2
 output: 
'''

def test_getTempDatSim():
    entry1 = dasEntry2
    entry2 = dasEntry4
    tempSimilarity = SimScore.getTempDatSim(entry1, entry2)
    print (tempSimilarity)
    ans = 0
    if (entry1["time"][0] == entry1["time"][0] and entry2["time"][0] == entry2["time"][0]):
        ans = 1
    elif (entry1["time"][0] != entry1["time"][0] and entry2["time"][0] != entry2["time"][0]):
        ans = 1
    assert ans == tempSimilarity

'''
 Function purpose: Tests whether the file types of two inputs are the same
 input: 
 output: 
'''
#TODO
# def test_DatatypeSim(dict, entry1, entry2):


'''
 Function purpose: Tests the inputs for size-similarity
 input: arrays, entry1, entry2 ?????????????
 output: boolean true/false
'''
def test_getGeoExtSim():
    entry1 = dasEntry2
    entry2 = dasEntry4
    GeoExtentSimilarity = SimScore.getGeoExtSim(entry1, entry2)
    print (GeoExtentSimilarity)
    ans = False
    if (GeoExtentSimilarity >= 0 and GeoExtentSimilarity <= 1):
        ans = True
    assert ans == True

'''
 Function purpose: Tests the inputs for time-similarity
 input: arrays, entry1, entry2 ?????????????
 output: boolean true/false
'''
def test_getTempExtSim():
    entry1 = dasEntry2
    entry2 = dasEntry4
    TemporalExtentSimilarity = SimScore.getTempExtSim(entry1, entry2)
    print (TemporalExtentSimilarity)
    ans = False
    if (TemporalExtentSimilarity > 0 and TemporalExtentSimilarity <= 1):
        ans = True
    assert ans == True


'''
 Function purpose: Calculates ratio between intersection of both bounding boxes and first zwei entries entry,
 0 if disjunct, given two entries as dicts
  Case distinction:
    1. Cut: Calculation of the cut surface of the bboxes with getAr, then ratio cut-surface / areaA
    2. A lies in B: Similarity 1
    3. B lies in A: Calculation Ratio AreaB / AreaA
    4. No cut: Similarity 0
 input: arrays, entry1, entry2 ?????????????
 output: boolean true/false
'''
def test_getInterGeoSim():
    entry1 = dasEntry2
    entry2 = dasEntry4
    InterGeoSimilarity = SimScore.getInterGeoSim(entry1, entry2)
    print (InterGeoSimilarity)
    minLatA=entry1["wkt_geometry"][0]
    maxLatA=entry1["wkt_geometry"][1]
    minLonA=entry1["wkt_geometry"][2]
    maxLonA=entry1["wkt_geometry"][3]
    minLatB=entry2["wkt_geometry"][0]
    maxLatB=entry2["wkt_geometry"][1]
    minLonB=entry2["wkt_geometry"][2]
    maxLonB=entry2["wkt_geometry"][3]

    # A in B
    # ans = False
    # The other one stands here only, so that no warning happens 
    ans = 0
    ans = ans +1
    if((minLonA > minLonB) and (maxLonA < maxLonB) and (minLatA > minLatB) and (maxLatA < maxLatB)):
        ans = True

#TODO


# def test_getInterTempSim(entry1, entry2):
   # InterTempSimilarity = SimScore.getInterTempSim(entry1, entry2)
   # print (InterTempSimilarity)
  

'''
 Function purpose: Tests whether the difference between the bounding boxes of Entry1 and Entry2 is the same
 as between Entry2 and Entry1
 input: arrays, entry1, entry2 ?????????????
 output: boolean true/false
'''
def test_getCenterGeoSim():
    entry1 = dasEntry2
    entry2 = dasEntry4
    geoCentSim1 = SimScore.getCenterGeoSim(entry1, entry2)
    geoCentSim2 = SimScore.getCenterGeoSim(entry2, entry1)
    print (geoCentSim1)
    print (geoCentSim2)    
    ans = False
    if (geoCentSim1 == geoCentSim2 and (geoCentSim1 >= 0 and geoCentSim1 <= 1)):
        ans = True
    assert ans == True   

'''
 Function purpose: Tests whether the difference between the Intervals of Entry1 and Entry2 is the same
 as between Entry2 and Entry1
 input: arrays, entry1, entry2 ?????????????
 output: boolean true/false
'''
def test_getCenterTempSim():
    entry1 = dasEntry2
    entry2 = dasEntry4
    tempCentSim1 = SimScore.getCenterTempSim(entry1, entry2)
    tempCentSim2 = SimScore.getCenterTempSim(entry2, entry1)
    print (tempCentSim1)
    print (tempCentSim2)    
    ans = False
    if (tempCentSim1 == tempCentSim2 and (tempCentSim1 >= 0 and tempCentSim1 <= 1)):
        ans = True
    assert ans == True   


# Helpfunction:

def intBetween(x, bottom, top):
    if(x>=bottom and x<=top and math.floor(x) == x):
        return True
    else:
        return False 

'''
 Function purpose: Tests whether the validity for getSimilarityScores inputs is correct
 input: dict, entries
 output: boolean true/false
'''

def test_checkValidity():
    entries = dasDict
    cmps = 5
    n = 1
    g = 4
    e = 3
    t = 4
    d = 2
    l = 0
    Validity = SimScore.checkValidity(entries, cmps, n, e, d, l, g, t)
    print(Validity)
    ans = False

    #entries a Dict
    if type (entries) is dict:
        #cmps ein int >= 0
        if (cmps >= 0 and math.floor(cmps) == cmps):
            #n ein int >= 1
            if (n >= 1 and math.floor(n) == n):
                #e,d,l,g,t >=0 and <=5
                loop = [e,d,l,g,t]
                temp = True
                for x in loop:
                    if intBetween(x,0,5) == False:
                        temp = False
                        break
                if(temp == True):
                    ans = True
    assert ans == True                    
               
'''
 Function purpose: Tests whether the diagonal length of bounding box (from entry as dict) is correct
 input: dict, entry
 output: boolean true/false
'''
def test_getDiagonal():
    entry = dasEntry2
    Diagonale = SimScore.getDiagonal(entry)
    print (Diagonale)
    # ans = False
    # The other one stands here only, so that no warning happens    
    ans = 0
    ans = ans +1 
    if ():
        ans = 9
        #TODO                                                                                                    

'''
 Function purpose: Tests whether the length of temporal interval (from entry as dict) is correct
 input: dict, entry
 output: boolean true/false
'''
def test_getInterv():
    entry = dasEntry2
    Intervall = SimScore.getInterv(entry)
    print (Intervall)
    ans = False
    if(Intervall == entry["time"][1] - entry ["time"][0]):
        ans = True
    assert ans == True
    
'''
 Function purpose: Tests whether the calculates area of bouunding box on earth surface is correct
 input: dict, entry
 output: boolean true/false
'''
def test_getAr():
    points = dasEntry2["wkt_geometry"]
    Ar = SimScore.getAr(points)
    print (Ar)
    ans = 1
    if ((points[0] == points[2]) and (points[1] == points[3])):
        ans = 0
    if(Ar == 0 and ans == 0):
        ans = True
    elif(Ar != 0 and ans != 0):
        ans = True
    else:
        ans = False
    assert ans == True




'''
@author: Benjamin Dietz
'''

import datetime, xarray
import gdal
from datetime import datetime as dtime
from netCDF4 import Dataset as NCDataset
import helpfunctions as hf
import convex_hull

DATATYPE = "application/x-netcdf"

def isValid(path):
    '''Checks whether it is valid netCDF or not. \n
    input "path": type string, path to file which shall be extracted \n
    output true if file is valid, false if not
    '''
    try:
        file = xarray.open_dataset(path)
        ncDataset = NCDataset(path)
        datasetGDAL = gdal.Open(path)
        if file is None or ncDataset is None or datasetGDAL is None:
            return False
    except:
        return False
    return True

def getVectorRepresentation(path):
    ''' abstracts the geometry of the file with a polygon
    first: collects all the points of the file
    then: call the function that computes the polygon of it \n
    input "path": type string, path to file which shall be extracted \n
    returns extracted coordinates of content: type list, list of lists with length = 2: type list
    '''
    file = xarray.open_dataset(path)
    if file is not None:
        if 'coords' in file.to_dict():
            if all (x in file.to_dict()["coords"] for x in ['lat', 'lon']):
                if 'data' in file.to_dict()["coords"]["lat"] and 'data' in file.to_dict()["coords"]["lon"]:
                    lats = file.to_dict()["coords"]["lat"]["data"]
                    lons = file.to_dict()["coords"]["lon"]["data"]
    if not ('lats' in locals() and 'lons' in locals()):
        ncDataset = NCDataset(path)
        if 'latitude' in ncDataset.variables:
            latitudes = ncDataset.variables["latitude"][:]
            lats = []
            for x in latitudes:
                lats.append(x)
        if 'longitude' in ncDataset.variables:
            longitudes = ncDataset.variables["longitude"][:]
            lons = []
            for x in longitudes:
                lons.append(x)
    if 'lats' in locals()  and 'lons' in locals():
        coordinates = { 'lat': lats, 'lon': lons }
        if len(lats) == len(lons):
            referenced_coordinates = []
            for index, val in enumerate(lats):
                referenced_coordinates.append([lons[index], val])
                coordinates = convex_hull.graham_scan(coordinates)
        else: 
            raise Exception("The numer auf lat-points and lon-points is not equal, thus a  vector representation cannot be computed.")
        return coordinates
    raise Exception("The vector representaton could not be extracted from the file")




def getBoundingBox(path):
    ''' extracts bounding box from NetCDF \n
    input "path": type string, path to file which shall be extracted \n
    returns bounding box: type list, length = 4 , type = float, schema = [min(longs), min(lats), max(longs), max(lats)] 
    '''
    ncDataset = NCDataset(path)
    if 'latitude' in ncDataset.variables:
        lats = ncDataset.variables["latitude"][:]
        if 'longitude' in ncDataset.variables:
            lngs = ncDataset.variables["longitude"][:]
            bbox =  [min(lngs), min(lats), max(lngs), max(lats)]
            return bbox
    if 'bbox' not in locals():
        # bbox not yet extracted
        xarrayForNetCDF = xarray.open_dataset(path)
        if xarrayForNetCDF is not None:
                if 'coords' in xarrayForNetCDF.to_dict():
                    if all (x in xarrayForNetCDF.to_dict()["coords"] for x in ['lat', 'lon']):
                        if xarrayForNetCDF.to_dict()["coords"]["lat"] is not None and xarrayForNetCDF.to_dict()["coords"]["lon"] is not None:
                            lats = xarrayForNetCDF.to_dict()["coords"]["lat"]
                            dataLats = lats["data"]
                            lons = xarrayForNetCDF.to_dict()["coords"]["lon"]
                            dataLons = lons["data"]
                            if len(dataLats) > 0 and len(dataLons) > 0:
                                bbox = [min(dataLons), min(dataLats), max(dataLons), max(dataLats)]
                            return bbox

    raise Exception("The bounding box could not be extracted from the file")




def getCRS(path):
    ''' gets the coordinate reference systems from the NetCDF file \n
    input "path": type string, path to file which shall be extracted \n
    returns epsg code of the used coordiante reference system: type list, list with two elements: 1. Crs of lats and 2. Crs of lons
    '''
    return hf.WGS84_EPSG_ID




def getTemporalExtent(path):
    ''' extracts the temporal extent of the netCDF file \n
    input "path": type string, path to file which shall be extracted \n
    reutrns temporal  extent of the file: type list, length = 2, both entries have the type dateTime, temporalExtent[0] <= temporalExtent[1]
    '''
    ncDataset = NCDataset(path)
    datasetGDAL = gdal.Open(path)
    metadataGDAL = datasetGDAL.GetMetadata()
    try:
        if 'time' in ncDataset.variables:
            times = ncDataset.variables["time"][:]
            temporal_extent =  str([min(times), max(times)]) + " Warning: Unit not absolute" 
            def getAbsoulteTimestamp(plusdays, steps, origin):
                origin = dtime.strptime(origin ,'%Y-%m-%d %H:%M:%S')
                if "days" in steps:
                    return origin + datetime.timedelta(days=plusdays)
                elif "hours" in steps:
                    return origin + datetime.timedelta(hours=plusdays)
                elif "minutes" in steps:
                    return origin + datetime.timedelta(minutes=plusdays)
                elif "seconds" in steps:
                    return origin + datetime.timedelta(seconds=plusdays)       
            if hasattr(ncDataset.variables["time"], 'units'):
                unit = ncDataset.variables["time"].units
                steps = unit[:unit.rfind(" since")]
                origin = unit[unit.rfind("since ")+6:]
                if origin[:4] == "0000":
                    origin = "2000" + origin[4:]
                if len(origin) < 11:
                    origin += " 00:00:00"

            elif "time#units" in metadataGDAL:
                    unit = metadataGDAL["time#units"]
                    steps = unit[:unit.rfind(" since")]
                    origin = unit[unit.rfind("since ")+6:]
            if times is not None:
                if len(times) > 0:
                    if 'steps' in locals():
                        temporal_extent = [str(getAbsoulteTimestamp(min(times), steps, origin)), str(getAbsoulteTimestamp(max(times), steps, origin))]
                        if getAbsoulteTimestamp(min(times), steps, origin) > datetime.datetime.now() or getAbsoulteTimestamp(max(times), steps, origin) > datetime.datetime.now():
                            print("temporal extent of " + path + " is not valid! (" + str(temporal_extent) + ")")
                        else:
                            return temporal_extent
    except:
        raise Exception("The temporal extent could not be extracted from the file")
    
    # raises exception when 1) no time variable could be found OR 2) no time unit could be found

    raise Exception("The temporal extent could not be extracted from the file")



'''
@author: Katharina Hovestadt
'''

import helpfunctions as hf
import json, gdal
from osgeo import ogr
import sys
from datetime import datetime
import datetime
from django.utils.dateparse import parse_datetime
import django, pytz
import unicodedata
import convex_hull
import pygeoj
import iso8601

DATATYPE = "application/geojson"

def extractContentFromPath(filePath):
    ''' method to extract geojson content from a file by using its filepath \n
    input "filepath": type string, path to file which shall be extracted \n
    returns geojson content of the filePath: type string,  returns  geojson content of filepath 
    '''  
    gjson = open(filePath, "rb")
    gjsonContent = json.load(gjson)
    gjson.close()
    return gjsonContent

   



def isValid(filePath):
    '''Checks whether it is valid geojson or not. \n
    input "filepath": type string, path to file which shall be extracted \n
    output true if file is valid, false if not
    '''
    try :   
        gjson = open(filePath, "rb")
        gjsonContent = json.load(gjson)
        gjson.close()
        if not gjsonContent:
            raise Exception('The geojson file from is empty')
        return True

    except ValueError as e:
        raise Exception ('The geojson file from ' + filePath + ' is not valid.' + str(e)) 

    except RuntimeError as e:
        raise Exception ('(geo)json file cannot be opened or read.' + str(e))
 
   
def convert3dto2d(filePath):
    '''transforms 3d to 2d coordinates in a geojson file. \n
    input "filepath": type string, path to file which shall be extracted \n
    returns geojson with 2d coordinates
    '''
    gjsonContent = extractContentFromPath(filePath)

    def extractAfterKeyword(searchParam, content):
        ''' searches for the value fo the dict entry with keyword which is given as input \n
        input "searchParam": type string, keyword for which is searched in the dict \n
        input "gjsonContent": type dict, Content of geojson File
        '''
        if type(content) == dict:
            for keyContent, valueContent in content.items():
                if keyContent == searchParam:   
                    valueContent = extractCoordinates(valueContent)
                if type(valueContent) == dict or type(valueContent) == list:
                    extractAfterKeyword(searchParam, valueContent)
        if type(content) == list:
            for element in content:
                extractAfterKeyword(searchParam, element)


    def extractCoordinates(coordsList):
        ''' extract coordinates out of a some more lists (e.g. with Multipolygons), cuts the height from 3d coordinates \n
        input "coordsList": type list, value of dict entry with key "coordinates" \n
        returns list with 2 coordinates or list with list with 2 coordinates
        '''
        if type(coordsList) == list and len(coordsList) == 3 and (type(coordsList[0]) == float or type(coordsList[0]) == int) and (type(coordsList[1]) == float or type(coordsList[1]) == int) and (type(coordsList[2]) == float or type(coordsList[2]) == int):
            coordsList = coordsList.pop()
        elif type(coordsList) == list and len(coordsList) != 0:
            for value in coordsList:
                extractCoordinates(value)
        return coordsList

    extractAfterKeyword("coordinates", gjsonContent)
    return gjsonContent


def getBoundingBox (filePath):
    ''' extract bounding box from geojson content \n
    input "filePath": type string, file path to geojson File \n
    returns bounding box: type list, length = 4 , type = float, schema = [min(longs), min(lats), max(longs), max(lats)] 
    '''
    bbox = None
    
    #gjsonContent is a FeatureCollection
    try:
        gjsonContent = pygeoj.load(data = convert3dto2d(filePath))
        bbox = gjsonContent.bbox
    #gjsonContent is a single geometrie and has to be converted to a FeatureCollection
    except ValueError:
        gjsonContent = convert3dto2d(filePath)

        gjsonFeatureColl = {"type": "FeatureCollection", "features": []}
        gjsonFeatureColl["features"].append(gjsonContent)
        gjsonContent_FeatureColl = pygeoj.load(data=gjsonFeatureColl)
        bbox = gjsonContent_FeatureColl.bbox
    
    if not bbox:
        raise Exception("Bounding box could not be extracted")
    return bbox




def getCRS(filePath):
    ''' extracts EPSG number of the taken coordinate reference system (short: crs), as standard the crs WGS84 is used. \n
    input "filePath": type string, file path to geojson File \n
    returns the epsg code of the used coordinate reference system: type int, EPSG number of taken crs
    ''' 
    
    def extractAfterKeyword(searchParam, gjsonContent):
        ''' searches for the value fo the dict entry with keyword which is given as input \n
        input "searchParam": type string, keyword for which is searched in the dict \n
        input "gjsonContent": type dict, Content of geojson File
        '''
        if type(gjsonContent) == dict:
            for keyContent, valueContent in gjsonContent.items():
                if keyContent == searchParam:   
                    extracted.append(valueContent)
                if type(valueContent) == dict or type(valueContent) == list:
                    extractAfterKeyword(searchParam, valueContent)
        if type(gjsonContent) == list:
            for element in gjsonContent:
                extractAfterKeyword(searchParam, element)


    try:
        gjsonContent = pygeoj.load(filePath)
        crsCode = gjsonContent.crs    
        if not crsCode:
            return hf.WGS84_EPSG_ID
        else: 
            for key, value in crsCode.items():
                if key == "properties":
                    try:
                        if value["name"] == "urn:ogc:def:crs:OGC:2:84":
                            return hf.WGS84_EPSG_ID
                        elif value["name"]:
                            splittedCrs = value["name"].split(":")
                            for elem in splittedCrs:
                                try:
                                    if int(elem) is not None:
                                        crsCode = int(elem)
                                        return crsCode
                                except:
                                    pass
                    except:
                        pass
                #formats like urn:ogc:def:crs:EPSG::25832
                
        return hf.WGS84_EPSG_ID
    except:
        gjsonContent = extractContentFromPath(filePath)
    
        #4326 is the standard epsg after http://wiki.geojson.org/GeoJSON_draft_version_6#Specification
        crsCode = hf.WGS84_EPSG_ID
        extracted = []
        extractAfterKeyword("crs", gjsonContent)
        if len(extracted) != 0:
            if type(extracted[0]) == dict and "properties" in extracted[0] and "code" in extracted[0]["properties"]:
                        crsCode = extracted[0]["properties"]["code"]
        return crsCode



 
def getVectorRepresentation(filePath):
    ''' extracts coordinates from geojson File (for vector representation) \n
    input "filePath": type string, file path to geojson File \n
    returns extracted coordinates of content: type list, list of lists with length = 2
    '''
    #type list, contains all coordinates as tuples
    coordinates = []

    #type list, contains everything extracted
    extracted = []




    def extractAfterKeyword(searchParam, gjsonContent):
        ''' searches for the value fo the dict entry with keyword which is given as input \n
        input "searchParam": type string, keyword for which is searched in the dict \n
        input "gjsonContent": type dict, Content of geojson File
        '''
        if type(gjsonContent) == dict:
            for keyContent, valueContent in gjsonContent.items():
                if keyContent == searchParam:   
                    extracted.append(valueContent)
                if type(valueContent) == dict or type(valueContent) == list:
                    extractAfterKeyword(searchParam, valueContent)
        if type(gjsonContent) == list:
            for element in gjsonContent:
                extractAfterKeyword(searchParam, element)




    def extractCoordinates(coordsList):
        ''' extract coordinates as tuples out of a some more lists (e.g. with Multipolygons) \n
        input "coordsList": type list, value of dict entry with key "coordinates"
        '''
        if type(coordsList) == list and len(coordsList) == 2 and (type(coordsList[0]) == float or type(coordsList[0]) == int) and (type(coordsList[1]) == float or type(coordsList[1]) == int):
            coordinates.append(coordsList)
        elif type(coordsList) == list and len(coordsList) != 0:
            for value in coordsList:
                extractCoordinates(value)




    gjsonContent = convert3dto2d(filePath)
    extracted = []
    extractAfterKeyword("coordinates", gjsonContent)
    extractCoordinates(extracted)
    if not coordinates:
        raise Exception("No coordinates found in File. Vector Representation could not be extracted.")
    #convex hull is formed out of coordinates
    coordinates = convex_hull.graham_scan(coordinates)
    return coordinates

    


def getTemporalExtent (filePath):
    ''' extract time extent from json string \n
    input "filePath": type string, file path to geojson File \n
    returns the temporal extent of the file: type list, length = 2, both entries have the type dateTime, temporalExtent[0] <= temporalExtent[1]
    '''
    #type list, contains all dates
    dateArray = []
    
    
    def searchForTimeElements(gjsonContent):
        ''' searches for time elements in a json \n
        input "gjsonContent": type dict, Content of geojson File
        '''
        
        ignore = ["created_at", "closed_at", "created", "closed", "initialize", "init", "last_viewed", "last_change", "change", "last_Change", "lastChange"] 
       

        if type(gjsonContent) == dict:
            for key, value in gjsonContent.items():     
                if key not in ignore:
                    searchForTimeElements(value)    
        elif type(gjsonContent) == list:
            for element in gjsonContent:
                searchForTimeElements(element)
        elif type(gjsonContent) == str:
            datetime_object = None
            datetime_object = parse_datetime(gjsonContent)
            if datetime_object == None:
                try:
                    datetime_object = iso8601.parse_date(gjsonContent)
                except:
                    pass
            if type(datetime_object) == datetime.datetime:
                dateArray.append(gjsonContent)



    gjsonContent = extractContentFromPath(filePath)
    temporalExtent = []

    searchForTimeElements(gjsonContent)
    if len(dateArray)!= 0:
        dateArray = sorted(dateArray)
        temporalExtent.append(dateArray[0])
        temporalExtent.append(dateArray[len(dateArray)-1])
    else:
        raise Exception("Could not extract timestamp.")
    return temporalExtent



import sys, os, getopt, datetime # important
from os import walk
import helpfunctions as hf
import handleShapefile, handleNetCDF, handleCSV, handleGeopackage, handleGeojson, handleXML, handleGML, handleKML
from lxml import etree
#handleGeotiff

COMMAND = None
XML_DIRPATH = None
CFG = None
RECURSIVE = False
OUTPUT_FILE = None
CSW_URL = None
XML = None
XSD = None
TIMEOUT = 30
FORCE_CONFIRM = False

#the capabilities of our CLI
def usage():
    """Provide usage instructions"""
    return '''
        NAME
            cli.py 
        SYNOPSIS
            cli.py -e </absoulte/path/to/record>|</absoulte/path/to/directory>
            cli.py -s </absoulte/path/to/record>|</absoulte/path/to/directory>
            cli.py -t </absoulte/path/to/record>|</absoulte/path/to/directory>
            Suppoted formats:
                    .dbf     |
                    .shp     |
                    .csv     |
                    .nc      |
                    .geojson |
                    .json    |
                    .gpkg    |
                    .geotiff |
                    .tif     |
                    .gml     |
            Available options:
            -e    Extract all metadata of a geospatial file
            -s    Extract all spatial metadata of a geospatial file
            -t    Extract all temporal metadata of a geospatial file
            -l    Load metadata of file on pycsw platform
            Possible metadata for formats:
                .shp and .dbf    :   bbox, shapetype, number of shape elements
                .csv             :   
                .nc              :   bbox (with differnet ellipsoid?), temporal extent (not always given)
                .geojson | .json :   
                .gpkg            :   bbox (with different ellipsoid?), temp extent still missing
                .geotiff | .tif  :   
                .gml             :   
            
            Available temporal metadata:
            
                [starting point, endpoint]
            
            Available spatial metadata:
                bbox
'''

def errorFunction():
    print("Error: A tag is required for a command")
    print(usage())

def noArgumentError():
    if len(sys.argv) == 1:
        print(usage())
        sys.exit(1)

def fehler():
    try:
        OPTS, ARGS = getopt.getopt(sys.argv[1:], 'e:s:t:l:')
    except getopt.GetoptError as err:
        print('\nERROR: %s' % err)
        print(usage())
        sys.exit(2)

def andereFehler():
    OPTS, ARGS = getopt.getopt(sys.argv[1:], 'e:s:t:l:')
    if len(OPTS) == 0:
        errorFunction()



def getDatabaseElementFromMetadata(metadataDictionary):
    ''' 
    still missing an NOT NULL: xml, anytext
    keys required in metadataDictionary: title, temporal_extent (in ISO standard?), fileformat (image/...)'''
    databaseElement = {}
# required metadata

    # identifier NOT NULL
    databaseElement["id    else: print(output)entifier"] = "?"
    databaseElement["typename"] = "csw:Record" # NOT NULL
    databaseElement["schema"] = "http://www.opengis.net/cat/csw/2.0.2" # NOT NULL
    databaseElement["mdsource"] = "local" # NOT NULL
    databaseElement["insert_date"] = str(datetime.datetime.now().replace(microsecond=0).isoformat()) + "Z" # https://stackoverflow.com/questions/2150739/iso-time-iso-8601-in-python NOT NULL
    if not all(x in databaseElement for x in ['typename', 'schema', "mdsource", "insert_date"]): # https://stackoverflow.com/questions/6159313/can-python-test-the-membership-of-multiple-values-in-a-list
        # metadata element cannot be saved
        raise Exception("One of the required metadata field could not be filled.")

# addictional metadata

    databaseElement["type"] = "http://purl.org/dc/dcmitype/Service" # ?

    # strict
    ''' only fills the value when search and found metadata fields are exactly the same '''
    searchStrict = ['title', 'date']
    for x in searchStrict:
        if x in metadataDictionary:
            databaseElement[x] = metadataDictionary[x]

    # non-strict
    ''' also fills the value if the searched parameter is part of a metadata fields in metadataDictionary '''
    searchNonStrict = ['format', 'source', 'crs', 'language', 'publisher', 'creator', 'resourcelanguage', 'contributor',
     'organization', 'securityconstraints', 'servicetype', 'servicetypeversion', 'links', 'degree', 'conditionapplyingtoaccessanduse',
     'title_alternate', 'abstract', 'keywords', 'keywordstype', 'relation', 'wkt_geometry', 'date_revision', 'date_creation', 'date_publication',
     'date_modified', 'specificationtitle', 'specificationdate', 'specificationdatetype' , 'otherconstraints']
    for x in searchNonStrict:
        if x in metadataDictionary:
            databaseElement[x] = metadataDictionary[x]
        else:
            for y in metadataDictionary:
                if x in y:
                    databaseElement[x] = str(metadataDictionary[y])
    
    # non-strict search for a different value as the key that gets entered in the databaseElement
    searchOtherWord = [[['constraint'], 'otherconstraints'], [['link'], 'links'], [['src'], 'source']]
    # schema of element in seachOtherWord: [[ <searchedWord1>, ... , <searchWordn> ], <keyToBeFilledInDatabase> ]
    for x in searchOtherWord:
        for key in metadataDictionary:
            for searchParam in x[0]:
                if searchParam in key:
                    databaseElement[x[1]] = str(metadataDictionary[key])

    
    # self extraction? -> many unclear metadata fields left
    # all of them (despite 'date') are already extracted when they are in the metadataDctionary. If we find better values here, 
    # we overwrite the old ones 
    if not 'language' in databaseElement:
        databaseElement["language"] = "?" # maybe get language from texts (title, history, description, ... ?)
        databaseElement["title_alternate"] = "?" # ?
        databaseElement["abstract"] = "?" # some text
        databaseElement["keywords"] = "?" # maybe from title?
    if 'keywords' in databaseElement:
        databaseElement["keywordstype"] = "?" # ?
        databaseElement["relation"] = "?" # ? some other UUID
        databaseElement["date"] = "?" # ?
        databaseElement["wkt_geometry"] = "?" # second abstraction level?
        databaseElement["date_revision"] = "?" # ?
        databaseElement["date_creation"] = "?" # ?
        databaseElement["date_publication"] = "?" # ?
        databaseElement["date_modified"] = "?" # ?
        databaseElement["specificationtitle"] = "?" # ?
        databaseElement["specificationdate"] = "?" # ?
    if 'specificationdate' in databaseElement:
        databaseElement["specificationdatetype"] = "?" # ?

    if 'temporal_extent' in metadataDictionary:
        if type(metadataDictionary["temporal_extent"]) == list and len(metadataDictionary["temporal_extent"]) == 2:
            databaseElement["time_begin"] = metadataDictionary["temporal_extent"][0] # in ISO standard?
            databaseElement["time_end"] = metadataDictionary["temporal_extent"][1] # in ISO standard?


    '''
        still some to come that does not really seem important
    '''
    def getXML():
        # required fileds in dataseElement: identifier, type,
        MY_NAMESPACES={'csw': 'http://www.opengis.net/cat/csw/2.0.2', 'ows': 'http://www.opengis.net/ows', 'dc': 'http://purl.org/dc/elements/1.1/', 'dct': 'http://purl.org/dc/terms/'}
        root = etree.Element("Record")
        root = etree.Element('{%s}Record' % MY_NAMESPACES['csw'], nsmap=MY_NAMESPACES)
        for x in databaseElement:
            child = etree.Element(('{%s}'+x) % MY_NAMESPACES['dc'])
            child.text = databaseElement[x]
            root.append(child)
        return root
    def getOnlyTextsFromXML(root):
        text = ""
        for index, child in enumerate(root):
            if index != len(root)-1:
                text += child.text + " "
        return text
    databaseElement["xml"] = etree.tostring(getXML(), encoding='utf8', method='xml', pretty_print=True)
    databaseElement["anytext"] = getOnlyTextsFromXML(getXML())
    if not all(x in databaseElement for x in ['anytext', "xml"]):
        # metadata element cannot be saved
        raise Exception("One of the required metadata field could not be filled.")
    else: hf.printObject(databaseElement)

# function is called when filePath is included in commanline (with tag 'e', 't' or 's')
# how this is done depends on the file format - the function calls the extractMetadataFrom<format>() - function
# returns False if the format is not supported, else returns True 
def extractMetadataFromFile(filePath, whatMetadata):
    a = filePath
    fileFormat = a[a.rfind('.')+1:]
    if fileFormat == 'shp' or fileFormat == 'dbf':
        metadata = handleShapefile.extractMetadata(fileFormat, filePath, whatMetadata)
    elif fileFormat == 'csv':
        metadata = handleCSV.extractMetadata(filePath, whatMetadata)
    elif fileFormat == 'nc':
        metadata = handleNetCDF.extractMetadata(fileFormat, filePath, whatMetadata)
    elif fileFormat == 'geojson' or fileFormat == 'json':
        metadata = handleGeojson.extractMetadataGeojson(fileFormat, filePath, whatMetadata)
    elif fileFormat == 'gpkg':
        metadata = handleGeopackage.extractMetadataGeopackage(filePath, whatMetadata)
    elif fileFormat == 'geotiff' or fileFormat == 'tif':
        metadata = handleGeotiff.extractMetadata(fileFormat, filePath, whatMetadata)
    elif fileFormat == 'gml':
        metadata = handleISO.extractMetadata(fileFormat, filePath, whatMetadata)
    else: raise Exception("This file format is not supported")
    print(metadata)
    #einheitliche Metadaten auf Crs prüfen, ggf umwandeln oder löschen, falls kein crs vorhanden.
    #{"bbox":[x1,y1,x2,y2],"title":"str", "time":[Start,End],"Crs":"Abkürzung"}
    return metadata


# function is called when path of directory is included in commanline (with tag 'e', 't' or 's')
def extractMetadataFromFolder(folderPath, whatMetadata):
    f = []
    for (dirpath, dirnames, filenames) in walk(folderPath):
        #fullPath = os.path.join(folderPath, filenames)
        f.extend(filenames)
        break
    metadataElements = []
    fullPaths = []
    for x in f:
        fullPaths.append(str(os.path.join(folderPath, x)))
    #handle each of the files in the folder seperate
    filesSkiped = 0
    for x in fullPaths:
        fileFormat = x[x.rfind('.')+1:]
        if fileFormat == 'shp': #here not 'dbf' so that it doesn take the object twice into account
            metadataElements.append(handleShapefile.extractMetadata("shp", x, whatMetadata))
        elif fileFormat == 'csv':
            metadataElements.append(handleCSV.extractMetadata(x, whatMetadata))
        elif fileFormat == 'nc':
            metadataElements.append(handleNetCDF.extractMetadata(fileFormat, x, whatMetadata))
        elif fileFormat == 'geojson' or fileFormat == 'json': #here both because it is either geojson OR json
            metadataElements.append(handleGeojson.extractMetadata(fileFormat, x, whatMetadata))
        elif fileFormat == 'gpkg':
            metadataElements.append(handleGeopackage.extractMetadata(x, whatMetadata))
        elif fileFormat == 'geotiff' or fileFormat == 'tif': #here both because it is either geotiff OR tif
            metadataElements.append(handleGeotiff.extractMetadata(fileFormat, x, whatMetadata))
        elif fileFormat == 'gml':
            metadataElements.append(handleISO.extractMetadata(fileFormat, x, whatMetadata))
        elif not (fileFormat == 'shp' or fileFormat == 'dbf'): 
            filesSkiped += 1
    if filesSkiped != 0: 
        print(str(filesSkiped) + ' file(s) has been skipped as its format is not suppoted; to see the suppoted formats look at -help')
    if len(metadataElements):
        hf.printObject(hf.extractCommonMetaDataOfMultiple(metadataElements, whatMetadata))
    else: print("No file in directory with metadata")

# tells the program what to do with certain tags and their attributes that are
# inserted over the command line
def nocheine():
    OPTS, ARGS = getopt.getopt(sys.argv[1:], 'e:s:t:l:')
    for o, a in OPTS:
        if o == '-e':
            COMMAND = a
            print("Extract all metadata:\n")
            if hf.exists(a):
                print("File exists")
                output = extractMetadataFromFile(a, 'e')

            elif os.path.isdir(a):
                #the input is a valid folder 
                extractMetadataFromFolder(a, 'e')
            else: print("\nFile or folder does not exist\n")
        elif o == '-t':
            print("\n")
            print("Extract Temporal metadata only:\n")
            COMMAND = a
            if hf.exists(a):
                output = extractMetadataFromFile(a, 't')
            elif os.path.isdir(a):
                #the input is a valid folder 
                extractMetadataFromFolder(a, 't')
            else: print("\nFile or folder does not exist\n")
        elif o == '-s':
            print("\n")
            print("Extract Spatial metadata only:\n")
            COMMAND = a
            if hf.exists(a):
                output = extractMetadataFromFile(a, 's')
            elif os.path.isdir(a):
                #the input is a valid folder 
                extractMetadataFromFolder(a, 's')
            else: print("\nFile or folder does not exist\n")
        elif o == '-l':
            print("\n")
            print("Load metadata of file on pycsw ...\n")
            COMMAND = a
            if hf.exists(a):
                output = getDatabaseElementFromMetadata(extractMetadataFromFile(a, 'e')) 
            elif os.path.isdir(a):
                #the input is a valid folder 
                raise Exception("Only single dictionaries can be uploaded into pycsw")
            else: print("\nFile or folder does not exist\n")
        elif o == '-help':
            print("\n")
            print(usage())
            print("\n")
        if type(output) == list or type(output) == dict:
            hf.printObject(output)
    else: print(output)

if __name__ == "__main__":
    fehler()
    andereFehler()
    nocheine()

extractMetadataFromFile("/home/ilka/Desktop/test.csv","s")


#!/usr/bin/env python
# coding: utf8
'''
@author: Benjamin Dietz, Niklas Aßelmann, Katharina Hovestadt, Ilka Pleiser
'''

import sys, os, platform, datetime, math, shapefile, fiona 
from datetime import datetime as dtime
from six.moves import configparser
from netCDF4 import Dataset as NCDataset
import netCDF4
import getopt
from os import walk
from osgeo import ogr
from osgeo import osr
from pyproj import Proj, transform
import convex_hull

WGS84_EPSG_ID = 4326

def printObject(object):
    '''
    Function purpose: output of metadata object \n
    Input: object \n
    Output: print("/n")
    '''
    print("\n")
    for a,b in object.items():
        print(str(a) + ": " + str(b))
    print("\n")




def exists(filename):
    '''
    Function purpose: checks whether the filename exists \n
    Input: filename \n
    Output: boolean
    '''
    if os.path.isfile(filename):
        return True
    else:
        return False




def computeBboxOfMultiple(bboxes):
    '''
    Function purpose:  
        compute an overall bbox from an array of bounding boxes
        required format of parameter 'bboxes': [ bbox1, bbox2, ..., bboxn ]
        while each bbox has the format: [ min(longs), min(lats), max(longs), max(lats) ]
    Input: bboxes\n
    Output: array of coordinates
    '''
    coordinate0 = 200
    coordinate1 = 200
    coordinate2 = -200
    coordinate3 = -200
    for x in bboxes:
        if x[0] < coordinate0:
            coordinate0 = x[0]
        if x[1] < coordinate1:
            coordinate1 = x[1]
        if x[2] > coordinate2:
            coordinate2 = x[2]
        if x[3] > coordinate3:
            coordinate3 = x[3]
    return [coordinate0, coordinate1, coordinate2, coordinate3]




def computeTempExtentOfMultiple(temporal_extents):
    '''
    Function purpose: 
        get multiple temporal extents in the schema [ temp1, temp2, ..., tempn ]
        with the schema of each temporal extent being: [ 'yyyy-mm-dd hh:mm:ss', 'yyyy-mm-dd hh:mm:ss' ]
        returning the overall temporal extent in ISO format
    Input: temporal_extents \n
    Output: array tempExtent (if startingpoint and endpoint follow given rules), None (if they don't)
    '''
    if len(temporal_extents) > 0:
        try:
            startPoint = dtime.strptime(temporal_extents[0][0] ,'%Y-%m-%dT%H:%M:%SZ')
            endPoint = dtime.strptime(temporal_extents[0][1] ,'%Y-%m-%dT%H:%M:%SZ')
            tempExtent = [startPoint, endPoint]
            for x in temporal_extents:
                startPoint = dtime.strptime(x[0] ,'%Y-%m-%dT%H:%M:%SZ')
                endPoint = dtime.strptime(x[1] ,'%Y-%m-%dT%H:%M:%SZ')
                if  startPoint < tempExtent[0]:
                    tempExtent = [ startPoint, tempExtent[1] ]
                if endPoint > tempExtent[1]:
                    tempExtent = [ tempExtent[0], endPoint ]
            return [tempExtent[0].isoformat(), tempExtent[1].isoformat()]
        except ValueError:
            try:
                startPoint = dtime.strptime(temporal_extents[0][0] ,'%Y-%m-%d %H:%M:%S')
                endPoint = dtime.strptime(temporal_extents[0][1] ,'%Y-%m-%d %H:%M:%S')
                tempExtent = [startPoint, endPoint]
                for x in temporal_extents:
                    startPoint = dtime.strptime(x[0] ,'%Y-%m-%d %H:%M:%S')
                    endPoint = dtime.strptime(x[1] ,'%Y-%m-%d %H:%M:%S')
                    if  startPoint < tempExtent[0]:
                        tempExtent = [ startPoint, tempExtent[1] ]
                    if endPoint > tempExtent[1]:
                        tempExtent = [ tempExtent[0], endPoint ]
                return [tempExtent[0].isoformat() + "Z", tempExtent[1].isoformat() + "Z"]
            except ValueError:
                startPoint = dtime.strptime(temporal_extents[0][0] ,'%Y-%m-%dT%H:%M:%S')
                endPoint = dtime.strptime(temporal_extents[0][1] ,'%Y-%m-%dT%H:%M:%S')
                tempExtent = [startPoint, endPoint]
                for x in temporal_extents:
                    startPoint = dtime.strptime(x[0] ,'%Y-%m-%dT%H:%M:%S')
                    endPoint = dtime.strptime(x[1] ,'%Y-%m-%dT%H:%M:%S')
                    if  startPoint < tempExtent[0]:
                        tempExtent = [ startPoint, tempExtent[1] ]
                    if endPoint > tempExtent[1]:
                        tempExtent = [ tempExtent[0], endPoint ]
                return [tempExtent[0].isoformat() + "Z", tempExtent[1].isoformat() + "Z"]

    else: return None




def countElements(array):
    '''
    Function purpose: return the occurancies of all values in the array \n
    Input: array \n
    Output: array list
    '''
    list = []
    for x in array:
            if [x, array.count(x)] not in list:
                list.append([x, array.count(x)])
    return list




def getAllRowElements(rowname,elements):
    '''
    Function purpose: help-function to get all row elements for a specific string \n
    Input: rowname, elements \n
    Output: array values
    '''
    for idx, val in enumerate(elements[0]):
        if  rowname in val:
            indexOf = idx
            values = []
            for x in elements:
                if x[indexOf] != rowname:
                    values.append(x[indexOf])
            return values




def searchForParameters(elements, paramArray):
    '''
    Function purpose: return all attributes of a elements in the first row of a file \n
    Input: paramArray, elements \n
    Output: getAllRowElements(x,elements)
    '''
    for x in paramArray:
        for row in elements[0]:
            if x in row:
                return getAllRowElements(x,elements)




def transformingIntoWGS84 (crs, coordinate):
    '''
    Function purpose: transforming SRS into WGS84 (EPSG:4978; used by the GPS satellite navigation system) \n
    Input: crs, point \n
    Output: retPoint constisting of x2, y2 (transformed points)
    '''
    source = osr.SpatialReference()
    source.ImportFromEPSG(int(crs))

    target = osr.SpatialReference()
    target.ImportFromEPSG(4326)

    transform = osr.CoordinateTransformation(source, target)
        
    point = ogr.Geometry(ogr.wkbPoint)
    point.AddPoint(float(coordinate[0]), float(coordinate[1]))
    point = point.ExportToWkt()
    point = ogr.CreateGeometryFromWkt(point)
  
    point.Transform(transform)
    return [point.GetX(), point.GetY()]




def disablePrint():
    ''' Disable Prints '''
    sys.stdout = open(os.devnull, 'w')




def enablePrint():
    ''' Restore Prints'''
    sys.stdout = sys.__stdout__




def transformingArrayIntoWGS84(crs, pointArray):
    '''
    Function purpose: transforming SRS into WGS84 (EPSG:4978; used by the GPS satellite navigation system) from an array \n
    Input: crs, pointArray \n
    Output: array array
    '''
    array = []
    ##vector_rep
    if type(pointArray[0]) == list:
        for x in pointArray:
            array.append(transformingIntoWGS84(crs, x))
        return array
    #bbox
    elif len(pointArray) == 4:
        bbox = [[pointArray[0], pointArray[1]],[pointArray[2], pointArray[3]]]
        transf_bbox = transformingArrayIntoWGS84(crs, bbox)
        return [transf_bbox[0][0],transf_bbox[0][1], transf_bbox[1][0], transf_bbox[1][1]]


'''
@author: Katharina Hovestadt
'''
import lxml.etree as etree
from owslib.util import http_post
import requests
from owslib.csw import CatalogueServiceWeb
import sys
import getopt
import subprocess
from urllib.request import urlopen
import os
from osgeo import ogr
import updateXml
sys.path.insert(1, os.path.join(sys.path[0], '..')+"/metadataExtraction") 
import extractFromFolderOrFile as extract
import configparser

COMMAND = None
XML_DIRPATH = None
CFG = None
RECURSIVE = False
OUTPUT_FILE = None
CSW_URL = None
XML = None
XSD = None
TIMEOUT = 30
FORCE_CONFIRM = False

# the capabilities of our CLI
def usage():
    """Provide usage instructions"""
    return '''
        NAME
            insertEntry.py 

        SYNOPSIS
            insertEntry.py -m </absoulte/path/to/metadataxml/of/record(s)> -s </absoulte/path/to/record>|</absoulte/path/to/directory/of/records> -p <url to server>

            p is optional - if it is not defined in the command, then the default value in the cunfiguration file (default_url.cfg) is taken as the url
'''


def errorFunction():
    print("Error: A tag is required for a command")
    print(usage())

if len(sys.argv) == 1:
    print(usage())
    sys.exit(1)

try:
    OPTS, ARGS = getopt.getopt(sys.argv[1:], 's:m:p:h:ho:')
except getopt.GetoptError as err:
    print('\nERROR: %s' % err)
    print(usage())
    #sys.exit(2)

if 'OPTS' in globals(): 
    if len(OPTS) == 0:
        errorFunction()


if 'OPTS' not in globals():
        raise Exception("The arguments -s \"[sourceOfFolderOrFile]\" and-m \"[sourceOfInserXMLFile]\" and -p \"[portOfServer]\" are required.")

source = None
insertXml = None
port = None

for o, a in OPTS:

    if o == '-s':
        source = a
        print(source)
        if "/" in a:
            ending = a[a.rfind("/")+1:]
        if '.' in ending:
            # handle it as a file
            metadata = extract.extractMetadataFromFile(a, 'e')
            if metadata is None:
                raise Exception("This file format is not supported")
        else:
            # handle it as a folder
            metadata = extract.extractMetadataFromFolder(a, 'e')
    elif o == '-m':
        insertXml = a
        print(insertXml)
    elif o == '-p':
        port = a
        print(port)
    elif o == '-h':
        print(usage())
    else: raise Exception("The argument " + o + " is not known. Please use -s \"[sourceOfFolderOrFile]\" and-m \"[sourceOfInsertXMLFile]\" and and -r \"[portOfServer]\".")

if not source:
    raise Exception("The argument -s is missing. Please give a source path of the used file/folder. Format: -s \"[sourceOfFileOrFolder]\"")
if not port:
    try:
        configParser = configparser.RawConfigParser()
        configFilePath = 'default_url.cfg'
        configParser.read(configFilePath)
        port = configParser.get('server', 'url')
    except:
        raise Exception("The argument -p is missing. Please give a url to the server. Format: -p \"[urlOfServer]\"")
if not insertXml:
    raise Exception("The argument -m is missing. Please give a source path to the insert xml file. Format:-m \"[sourceOfInsertXMLFile]\"  ")



xmlContent = open(insertXml)
tree = etree.parse(xmlContent)
root = tree.getroot()
#insert with typename "md:MD_Metadata"
if(root.tag == "{http://www.opengis.net/cat/csw/2.0.2}Transaction"):
    uuid = root[0][0][0][0].text
    try:
        #sends transaction-insert request as xml to pycsw server
        insert = requests.post(port, data = open(insertXml))
        print(insert.content)
    except Exception as e:
        print("Error while inserting db entry. "+ str(e))

#insert Record with typename csw:Record
else:
    uuid = root[0].text
    try:
        # sends transaction-insert request as xml to pycsw server
        csw = CatalogueServiceWeb(port)
        csw.transaction(ttype='insert', typename='csw:Record', record = xmlContent)
        insert = requests.post(port, data = open(insertXml))
        print(insert.content)
    except Exception as e:
        print("Error while inserting db entry. "+ str(e))

try:
    # sends transaction-update request as xml to pycsw server 
    update = requests.post(port, data = updateXml.createXmlTree(metadata, uuid))
    print(update.content)
except Exception as e:
    print("Error while updating db entry with id "+ uuid +". "+ str(e))






'''
@author: Benjamin Dietz
'''

import shapefile, fiona
import helpfunctions as hf
import gdal
from osgeo import ogr
import sys
import convex_hull

DATATYPE = "application/shp"

def isValid(path):
    '''Checks whether it is valid shapefile or not. \n
    input "path": type string, path to file which shall be extracted \n
    output true if file is valid, false if not
    '''
    pathWithoutEnding = path[:len(path)-4]
    if not (hf.exists(pathWithoutEnding + ".dbf") and hf.exists(pathWithoutEnding + ".shp") and \
        hf.exists(pathWithoutEnding + ".shx")):
        return False
    try:
        mydbf = open(pathWithoutEnding + ".dbf", "rb")
        myshp = open(pathWithoutEnding + ".shp", "rb")
        myshx = open(pathWithoutEnding + ".shx", "rb")
        r = shapefile.Reader(shp=myshp, dbf=mydbf, shx=myshx)
    except:
        return False
    return True

def getCRS(path):
    ''' gets the coordinate reference systems from the shapefile \n
    input "path": type string, file path to shapefile \n
    returns epsg code of the used coordinate reference system
    '''
    try:
        with fiona.open(path) as datasetFiona:
            if hasattr(datasetFiona, "crs"):
                if 'init' in datasetFiona.crs:
                    initField = datasetFiona.crs["init"]
                    crs = initField[initField.rfind(":") + 1 : ]
                    return int(crs)                
    except Exception as e:
        pathWithoutEnding = path[:len(path)-4]
        if not (hf.exists(pathWithoutEnding + ".dbf") and hf.exists(pathWithoutEnding + ".shp") and \
            hf.exists(pathWithoutEnding + ".shx")):
            raise Exception("One of the required files with the following ending are missing: .dbf, .shp or .shx")
        else:
            raise e
    raise Exception("The CRS cannot be extracted from shapefiles")


def getTemporalExtent(path):
    ''' extracts temporal extent of the shapefile \n
    input "path": type string, file path to shapefile file
    '''

    raise Exception("The temporal extent cannot be extracted of a shapefile")




def getVectorRepresentation(path):
    ''' abstract the geometry of the file with a polygon
    first: collects all the points of the file
    then: call the function that computes the polygon of it \n
    input "path": type string, file path to shapefile \n
    returns extracted coordinates of content from shapefiletype list, list of lists with length = 2, 
    '''
    try:
        if not '.shp' in path:
            shpPath = path[:path.rfind(".")+1]
            shpPath += "shp"
            if not hf.exists(shpPath):
                raise FileNotFoundError("Related shp-file could not be found!")
            else:
                path = shpPath
        with fiona.open(path) as datasetFiona:
            if datasetFiona is not None:
                coordinates = ""
                for x in datasetFiona:
                    if 'geometry' in x:
                        if 'coordinates' in x["geometry"]:
                            if len(x["geometry"]["coordinates"]) > 0:
                                coors = str(x["geometry"]["coordinates"][0])
                                coordinatesASString = coors[coors.find("(") : coors.rfind(")") + 1]
                                if len(coordinatesASString) < 3 and len(x["geometry"]["coordinates"]) == 2:
                                    coordinates += "(" + str(x["geometry"]["coordinates"][0]) + ", " + str(x["geometry"]["coordinates"][1]) + ")"
                                coordinates += coordinatesASString + ", "
                coordinates = coordinates[: len(coordinates) - 4]
                coordinates = coordinates.split("), (")
                for index, value in enumerate(coordinates):
                    coordinates[index] = str(value).split(", ")
            for index, value in enumerate(coordinates):
                if len(value) != 2:
                    print("Error: Coordinate does not have two values")
                try:
                    coordinates[index][0] = float(value[0].replace("(", "").replace(")", ""))
                    coordinates[index][1] = float(value[1].replace("(", "").replace(")", ""))
                except:
                    print("Error: Value cannot be converted into float" + value[0])
            coordinates = convex_hull.graham_scan(coordinates)
            return coordinates
    except Exception as e:
        pathWithoutEnding = path[:len(path)-4]
        if not (hf.exists(pathWithoutEnding + ".dbf") and hf.exists(pathWithoutEnding + ".shp") and \
            hf.exists(pathWithoutEnding + ".shx")):
            raise Exception("One of the required files with the following ending are missing: .dbf, .shp or .shx")
        else:
            raise e
    raise Exception("The vector representaton could not be extracted from the file")




def getBoundingBox(path):
    ''' extracts bounding box from shapfile \n
    input "path": type string, file path to shapefile \n
    returns bounding box of the file: type list, length = 4 , type = float, schema = [min(longs), min(lats), max(longs), max(lats)] 
    '''
    # try to get the bounding box with fiona
    try:
        with fiona.open(path) as datasetFiona:
            if hasattr(datasetFiona, "bounds"):
                if len(datasetFiona.bounds) > 3:
                    bboxInOriginalCRS = [datasetFiona.bounds[0], datasetFiona.bounds[1], datasetFiona.bounds[2], datasetFiona.bounds[3]]
                    return bboxInOriginalCRS

        # if fiona is not working (on this file), try to get the bbox with the module 'shapefile'
        pathWithoutEnding = path[:len(path)-4]
        if '.shp' in path:
            if hf.exists(pathWithoutEnding + ".dbf"):
                mydbf = open(pathWithoutEnding + ".dbf", "rb")
                myshp = open(path, "rb")
        elif '.dbf' in path:
            if hf.exists(pathWithoutEnding + ".shp"):
                myshp = open(pathWithoutEnding + ".shp", "rb")
                mydbf = open(path, "rb")
        if 'myshp' in locals():
            if 'mydbf' in locals():
                r = shapefile.Reader(shp=myshp, dbf=mydbf)
                boundingbox = r.bbox
                if type(boundingbox) == list:
                    if len(boundingbox) == 6:
                        boundingbox.pop(2)
                        boundingbox.pop(len(boundingbox)-1)
                return boundingbox
    except Exception as e:
        pathWithoutEnding = path[:len(path)-4]
        if not (hf.exists(pathWithoutEnding + ".dbf") and hf.exists(pathWithoutEnding + ".shp") and \
            hf.exists(pathWithoutEnding + ".shx")):
            raise Exception("One of the required files with the following ending are missing: .dbf, .shp or .shx")
        else:
            raise e

    raise Exception("The bounding box could not be extracted from the file")
